(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/proxy.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
var ChildProcess = Npm.require("child_process");                       // 18
var Fs = Npm.require("fs");                                            // 19
var Path = Npm.require("path");                                        // 20
var Future = Npm.require("fibers/future");                             // 21
var Http = Npm.require("http");                                        // 22
var Url = Npm.require("url");                                          // 23
var Promise = Npm.require("es6-promise").Promise;                      // 24
var Capnp = Npm.require("capnp");                                      // 25
var Net = Npm.require("net");                                          // 26
                                                                       //
var ByteStream = Capnp.importSystem("sandstorm/util.capnp").ByteStream;
var ApiSession = Capnp.importSystem("sandstorm/api-session.capnp").ApiSession;
var WebSession = Capnp.importSystem("sandstorm/web-session.capnp").WebSession;
var HackSession = Capnp.importSystem("sandstorm/hack-session.capnp");  // 31
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;   // 33
                                                                       //
SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;           // 35
SANDSTORM_LOGDIR = (SANDSTORM_ALTHOME || "") + "/var/log";             // 36
SANDSTORM_VARDIR = (SANDSTORM_ALTHOME || "") + "/var/sandstorm";       // 37
                                                                       //
var storeReferralProgramInfoApiTokenCreated = function (db, accountId, identityId, apiTokenAccountId) {
  // From the Referral program's perspective, if Bob's Account has no referredByComplete, then we
  // update Bob's Identity to say it's referredBy Alice's Account (which is apiTokenAccountId).
  check(accountId, String);                                            // 42
  check(identityId, String);                                           // 43
  check(apiTokenAccountId, String);                                    // 44
                                                                       //
  // Bail out early if quota enforcement is disabled.                  //
  if (!Meteor.settings["public"].quotaEnabled) {                       // 47
    return;                                                            // 48
  }                                                                    //
                                                                       //
  var aliceAccountId = apiTokenAccountId;                              // 51
  var bobAccountId = accountId;                                        // 52
  var bobIdentityId = identityId;                                      // 53
                                                                       //
  if (Meteor.users.find({                                              // 55
    _id: bobAccountId,                                                 // 56
    referredByComplete: { $exists: true }                              // 57
  }).count() > 0) {                                                    //
    return;                                                            // 59
  }                                                                    //
                                                                       //
  // Only actually update Bob's Identity ID if there is no referredBy.
  Meteor.users.update({ _id: bobIdentityId, referredBy: { $exists: false } }, { $set: { referredBy: aliceAccountId } });
};                                                                     //
                                                                       //
function referralProgramLogSharingTokenUse(db, bobAccountId) {         // 68
  // Hooray! The sharing token is valid! Someone (let's call them Charlie) is going to get a UiView
  // to this grain!  This means that the user who created this apiToken knows how to use the 'share
  // access' interface. Let's call them Bob.                           //
  //                                                                   //
  // If Bob's Account.referredByComplete is not yet set, then look at Bob's Identities and take the
  // first referredBy we find -- let's call that Alice.                //
  //                                                                   //
  // We copy Alice's account ID to Bob's Account.referredByComplete, and then update Alice's
  // referredIdentityIds to point at Bob's Identity, and then remove the referredBy from Bob's
  // Identity since it has become redundant.                           //
  //                                                                   //
  // Implementation note: this does mean that Alice can get referral credit for Bob by sharing a
  // link with Bob, even if Bob already had an account.                //
                                                                       //
  // Bail out early if quota support is not enabled.                   //
  if (!Meteor.settings["public"].quotaEnabled) {                       // 84
    return;                                                            // 85
  }                                                                    //
                                                                       //
  // Bail out if Bob has a referredByComplete.                         //
  if (Meteor.users.find({ _id: bobAccountId, referredByComplete: { $exists: true } }).count() > 0) {
    return;                                                            // 90
  }                                                                    //
                                                                       //
  // Look for a referredBy on any of Bob's identities.                 //
  var bobIdentityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({ _id: bobAccountId }));
  var bobIdentityWithReferredBy = Meteor.users.findOne({               // 95
    _id: { $in: bobIdentityIds },                                      // 96
    referredBy: { $exists: true }                                      // 97
  });                                                                  //
                                                                       //
  if (!bobIdentityWithReferredBy) {                                    // 100
    return;                                                            // 101
  }                                                                    //
                                                                       //
  var aliceAccountId = bobIdentityWithReferredBy.referredBy;           // 104
                                                                       //
  // Store Bob's Account.referralCompletedBy.                          //
  var now = new Date();                                                // 107
  Meteor.users.update({                                                // 108
    _id: bobAccountId,                                                 // 109
    referredByComplete: { $exists: false }                             // 110
  }, {                                                                 //
    $set: {                                                            // 112
      referredByComplete: bobIdentityWithReferredBy.referredBy,        // 113
      referredCompleteDate: now                                        // 114
    }                                                                  //
  });                                                                  //
                                                                       //
  // Update Alice's Account.referredIdentityIds.                       //
  Meteor.users.update({ _id: aliceAccountId }, {                       // 119
    $push: { referredIdentityIds: bobIdentityWithReferredBy._id }      // 120
  });                                                                  //
                                                                       //
  // Remove now-redundant Bob identity referredBy.                     //
  Meteor.users.update({ _id: bobIdentityWithReferredBy._id }, {        // 124
    $unset: { referredBy: true }                                       // 125
  });                                                                  //
}                                                                      //
                                                                       //
// User-agent strings that should be allowed to use http basic authentication.
// These are regex matches, so ensure they are escaped properly with double
// backslashes. For security reasons, we MUST NOT whitelist any user-agents
// that may render html and execute embedded scripts.                  //
BASIC_AUTH_USER_AGENTS = ["git\\/", "GitHub-Hookshot\\/", "mirall\\/", "Mozilla\\/5\\.0 \\([^\\\\]*\\) mirall\\/", "Mozilla\\/5\\.0 \\(iOS\\) ownCloud-iOS\\/", "Mozilla\\/5\\.0 \\(Android\\) ownCloud-android\\/", "litmus\\/"];
BASIC_AUTH_USER_AGENTS_REGEX = new RegExp("^(" + BASIC_AUTH_USER_AGENTS.join("|") + ")", "");
                                                                       //
var SESSION_PROXY_TIMEOUT = 60000;                                     // 144
                                                                       //
var sandstormCoreFactory = makeSandstormCoreFactory();                 // 146
var backendAddress = "unix:" + (SANDSTORM_ALTHOME || "") + Backend.socketPath;
var sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
var sandstormBackend = sandstormBackendConnection.restore(null, Backend);
                                                                       //
// TODO(cleanup): This initilization belongs with the rest of our package initialization in
//   db-deprecated.js. We can't put it there now because we need to contruct sandstormCoreFactory first.
globalBackend = new SandstormBackend(globalDb, sandstormBackend);      // 153
Meteor.onConnection(function (connection) {                            // 154
  connection.sandstormBackend = globalBackend;                         // 155
});                                                                    //
                                                                       //
// We've observed a problem in production where occasionally the front-end stops talking to the
// back-end. It happens very rarely -- like once a month -- and we've been unable to reproduce it
// in testing, making it very hard to debug. The problem appears both on Sandstorm and Blackrock.
// Restarting the node process (and only the node process) always fixes the problem.
//                                                                     //
// Here, I've added some code that attempts to detect the problem by doing a health check
// periodically and then remaking the connection if it seems broken. We'll see if this helps!
var backendHealthy = true;                                             // 165
var disconnectCount = 0;                                               // 166
Meteor.setInterval(function () {                                       // 167
  if (!backendHealthy) {                                               // 168
    if (disconnectCount++ > 2) process.abort();                        // 169
    console.error("error: Backend hasn't responded in 30 seconds! Reconnecting.");
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(true);
    sandstormBackendConnection.close();                                // 172
    sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
    sandstormBackend = sandstormBackendConnection.restore(null, Backend);
    globalBackend._backendCap = sandstormBackend;                      // 175
  }                                                                    //
                                                                       //
  var debugLog = !backendHealthy;                                      // 178
  backendHealthy = false;                                              // 179
  var promise = sandstormBackend.ping().then(function () {             // 180
    backendHealthy = true;                                             // 181
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(false);
  }, function (err) {                                                  //
    console.error("error: Backend ping threw error!", err.stack);      // 184
    // The connection will be remade on the next interval. Note that we do NOT normally observe
    // exceptions being thrown for this problem; we see the connection simply stop responding.
    // So we don't expect this branch to execute in any case.          //
  });                                                                  //
                                                                       //
  if (debugLog) {                                                      // 190
    console.log("capnp.js: outer promise:", promise);                  // 191
  }                                                                    //
}, 30000);                                                             //
                                                                       //
// =======================================================================================
// Meteor context <-> Async Node.js context adapters                   //
// TODO(cleanup):  Move to a different file.                           //
                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {    // 199
  callback();                                                          // 200
});                                                                    //
                                                                       //
inMeteor = function (callback) {                                       // 203
  // Calls the callback in a Meteor context.  Returns a Promise for its result.
  return new Promise(function (resolve, reject) {                      // 205
    inMeteorInternal(function () {                                     // 206
      try {                                                            // 207
        resolve(callback());                                           // 208
      } catch (err) {                                                  //
        reject(err);                                                   // 210
      }                                                                //
    });                                                                //
  });                                                                  //
};                                                                     //
                                                                       //
promiseToFuture = function (promise) {                                 // 216
  var result = new Future();                                           // 217
  promise.then(result["return"].bind(result), result["throw"].bind(result));
  return result;                                                       // 219
};                                                                     //
                                                                       //
waitPromise = function (promise) {                                     // 222
  return promiseToFuture(promise).wait();                              // 223
};                                                                     //
                                                                       //
// =======================================================================================
// API for creating / starting grains from Meteor methods.             //
                                                                       //
var proxiesByHostId = {};                                              // 229
                                                                       //
Meteor.methods({                                                       // 231
  newGrain: function (packageId, command, title, identityId) {         // 232
    // Create and start a new grain.                                   //
                                                                       //
    check(packageId, String);                                          // 235
    check(command, Object); // Manifest.Command from package.capnp.    // 236
    check(title, String);                                              // 237
    check(identityId, String);                                         // 238
                                                                       //
    if (!this.userId) {                                                // 240
      throw new Meteor.Error(403, "Unauthorized", "Must be logged in to create grains.");
    }                                                                  //
                                                                       //
    if (!globalDb.userHasIdentity(this.userId, identityId)) {          // 244
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    if (!isSignedUpOrDemo()) {                                         // 248
      throw new Meteor.Error(403, "Unauthorized", "Only invited users or demo users can create grains.");
    }                                                                  //
                                                                       //
    if (isUserOverQuota(Meteor.user())) {                              // 253
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");
    }                                                                  //
                                                                       //
    var pkg = Packages.findOne(packageId);                             // 258
    var isDev = false;                                                 // 259
    if (!pkg) {                                                        // 260
      // Maybe they wanted a dev package.  Check there too.            //
      pkg = DevPackages.findOne(packageId);                            // 262
      isDev = true;                                                    // 263
    }                                                                  //
                                                                       //
    if (!pkg) {                                                        // 266
      throw new Meteor.Error(404, "Not Found", "No such package is installed.");
    }                                                                  //
                                                                       //
    var appId = pkg.appId;                                             // 270
    var manifest = pkg.manifest;                                       // 271
    var grainId = Random.id(22); // 128 bits of entropy                // 272
    Grains.insert({                                                    // 273
      _id: grainId,                                                    // 274
      packageId: packageId,                                            // 275
      appId: appId,                                                    // 276
      appVersion: manifest.appVersion,                                 // 277
      userId: this.userId,                                             // 278
      identityId: identityId,                                          // 279
      title: title,                                                    // 280
      "private": true                                                  // 281
    });                                                                //
                                                                       //
    globalBackend.startGrainInternal(packageId, grainId, this.userId, command, true, isDev);
    globalBackend.updateLastActive(grainId, this.userId, identityId);  // 285
    return grainId;                                                    // 286
  },                                                                   //
                                                                       //
  openSession: function (grainId, identityId, cachedSalt) {            // 289
    // Open a new UI session on an existing grain.  Starts the grain if it is not already
    // running.                                                        //
                                                                       //
    check(grainId, String);                                            // 293
    check(identityId, Match.OneOf(undefined, null, String));           // 294
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 295
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    var db = this.connection.sandstormDb;                              // 301
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 302
    if (!SandstormPermissions.mayOpenGrain(db, { grain: { _id: grainId, identityId: identityId } })) {
      throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
    }                                                                  //
                                                                       //
    var opened = globalBackend.openSessionInternal(grainId, this.userId, identityId, null, null, cachedSalt);
    var result = opened.methodResult;                                  // 310
    var proxy = new Proxy(grainId, this.userId, result.sessionId, result.hostId, identityId, false, opened.supervisor);
    proxiesByHostId[result.hostId] = proxy;                            // 313
    return result;                                                     // 314
  },                                                                   //
                                                                       //
  openSessionFromApiToken: function (params, identityId, cachedSalt) {
    // Given an API token, either opens a new WebSession to the underlying grain or returns a
    // path to which the client should redirect in order to open such a session.
                                                                       //
    check(params, {                                                    // 321
      token: String,                                                   // 322
      incognito: Boolean                                               // 323
    });                                                                //
    check(identityId, Match.OneOf(undefined, null, String));           // 325
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 326
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);
    }                                                                  //
                                                                       //
    var token = params.token;                                          // 332
    var incognito = params.incognito;                                  // 333
    var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
    var apiToken = ApiTokens.findOne(hashedToken);                     // 335
    validateWebkey(apiToken);                                          // 336
    var grain = Grains.findOne({ _id: apiToken.grainId });             // 337
    if (!grain) {                                                      // 338
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + apiToken.grainId);
    }                                                                  //
                                                                       //
    if (apiToken.accountId) {                                          // 342
      referralProgramLogSharingTokenUse(globalDb, apiToken.accountId);
    }                                                                  //
                                                                       //
    var title = undefined;                                             // 346
    if (grain.userId === apiToken.accountId) {                         // 347
      title = grain.title;                                             // 348
    } else {                                                           //
      var sharerToken = apiToken.identityId && ApiTokens.findOne({     // 350
        grainId: apiToken.grainId,                                     // 352
        "owner.user.identityId": apiToken.identityId                   // 353
      }, {                                                             //
        sort: {                                                        // 355
          "owner.user.lastUsed": -1                                    // 356
        }                                                              //
      });                                                              //
      if (sharerToken) {                                               // 359
        title = sharerToken.owner.user.title;                          // 360
      } else {                                                         //
        title = "shared grain";                                        // 362
      }                                                                //
    }                                                                  //
                                                                       //
    if (this.userId && !incognito) {                                   // 366
      if (identityId != apiToken.identityId && identityId != grain.identityId && !ApiTokens.findOne({ "owner.user.identityId": identityId, parentToken: hashedToken })) {
        var owner = { user: { identityId: identityId, title: title } };
                                                                       //
        // Create a new API token for the identity redeeming this token.
        var result = SandstormPermissions.createNewApiToken(globalDb, { rawParentToken: token }, apiToken.grainId, apiToken.petname, { allAccess: null }, owner);
        globalDb.addContact(apiToken.accountId, identityId);           // 374
                                                                       //
        // If the parent API token is forSharing and it has an accountId, then the logged-in user (call
        // them Bob) is about to access a grain owned by someone (call them Alice) and save a reference
        // to it as a new ApiToken. (For share-by-link, this occurs when viewing the grain. For
        // share-by-identity, this happens immediately.)               //
        if (result.parentApiToken) {                                   // 380
          var parentApiToken = result.parentApiToken;                  // 381
          if (parentApiToken.forSharing && parentApiToken.accountId) {
            storeReferralProgramInfoApiTokenCreated(globalDb, this.userId, owner.user.identityId, parentApiToken.accountId);
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      return { redirectToGrain: apiToken.grainId };                    // 389
    } else {                                                           //
      if (!SandstormPermissions.mayOpenGrain(globalDb, { token: apiToken })) {
        throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
      }                                                                //
                                                                       //
      var opened = globalBackend.openSessionInternal(apiToken.grainId, null, null, title, apiToken, cachedSalt);
                                                                       //
      var result = opened.methodResult;                                // 399
      var proxy = new Proxy(apiToken.grainId, grain.userId, result.sessionId, result.hostId, identityId, false);
      proxy.apiToken = apiToken;                                       // 402
      proxiesByHostId[result.hostId] = proxy;                          // 403
      return result;                                                   // 404
    }                                                                  //
  },                                                                   //
                                                                       //
  keepSessionAlive: function (sessionId) {                             // 408
    // TODO(security):  Prevent draining someone else's quota by holding open several grains shared
    //   by them.                                                      //
    check(sessionId, String);                                          // 411
                                                                       //
    var session = Sessions.findAndModify({                             // 413
      query: { _id: sessionId },                                       // 414
      update: { $set: { timestamp: new Date().getTime() } },           // 415
      fields: { grainId: 1, identityId: 1, hostId: 1 }                 // 416
    });                                                                //
                                                                       //
    if (session) {                                                     // 419
      // Session still present in database, so send keep-alive to backend.
      try {                                                            // 421
        var grainId = session.grainId;                                 // 422
        var hostId = session.hostId;                                   // 423
        var supervisor = proxiesByHostId[hostId] && proxiesByHostId[hostId].supervisor;
        if (!supervisor) {                                             // 425
          supervisor = globalBackend.continueGrain(grainId).supervisor;
        }                                                              //
                                                                       //
        waitPromise(supervisor.keepAlive());                           // 429
        globalBackend.updateLastActive(grainId, this.userId, session.identityId);
      } catch (err) {                                                  //
        // Ignore disconnects, which imply that the grain shut down already. It'll start back up on
        // the next request, so whatever.                              //
        if (err.kjType !== "disconnected") {                           // 434
          throw err;                                                   // 435
        }                                                              //
      }                                                                //
                                                                       //
      return true;                                                     // 439
    } else {                                                           //
      return false;                                                    // 441
    }                                                                  //
  },                                                                   //
                                                                       //
  shutdownGrain: function (grainId) {                                  // 445
    check(grainId, String);                                            // 446
    var grain = Grains.findOne(grainId);                               // 447
    if (!grain || !this.userId || grain.userId !== this.userId) {      // 448
      throw new Meteor.Error(403, "Unauthorized", "User is not the owner of this grain");
    }                                                                  //
                                                                       //
    waitPromise(globalBackend.shutdownGrain(grainId, grain.userId, true));
  }                                                                    //
});                                                                    //
                                                                       //
var validateWebkey = function (apiToken, refreshedExpiration) {        // 456
  // Validates that `apiToken` is a valid UiView webkey, throwing an exception if it is not. If
  // `refreshedExpiration` is set and if the token has an `expiresIfUnused` field, then the
  // `expiresIfUnused` field is reset to `refreshedExpiration`.        //
                                                                       //
  if (!apiToken) {                                                     // 461
    throw new Meteor.Error(403, "Invalid authorization token");        // 462
  }                                                                    //
                                                                       //
  if (apiToken.revoked) {                                              // 465
    throw new Meteor.Error(403, "Authorization token has been revoked");
  }                                                                    //
                                                                       //
  if (apiToken.owner && !("webkey" in apiToken.owner)) {               // 469
    throw new Meteor.Error(403, "Unauthorized to open non-webkey token.");
  }                                                                    //
                                                                       //
  if (apiToken.expires && apiToken.expires.getTime() <= Date.now()) {  // 473
    throw new Meteor.Error(403, "Authorization token expired");        // 474
  }                                                                    //
                                                                       //
  if (apiToken.expiresIfUnused) {                                      // 477
    if (apiToken.expiresIfUnused.getTime() <= Date.now()) {            // 478
      throw new Meteor.Error(403, "Authorization token expired");      // 479
    } else if (refreshedExpiration) {                                  //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: refreshedExpiration } });
    } else {                                                           //
      // It's getting used now, so clear the expiresIfUnused field.    //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: null } });
    }                                                                  //
  }                                                                    //
                                                                       //
  if (apiToken.objectId || apiToken.frontendRef) {                     // 488
    throw new Meteor.Error(403, "ApiToken refers to a non-webview Capability.");
  }                                                                    //
};                                                                     //
                                                                       //
// Used by shared/grain.js (which sounds like a broken dependency, since this code is only available
// on the server)                                                      //
getGrainSize = function (supervisor, oldSize) {                        // 495
  var promise = undefined;                                             // 496
  if (oldSize === undefined) {                                         // 497
    promise = supervisor.getGrainSize();                               // 498
  } else {                                                             //
    promise = supervisor.getGrainSizeWhenDifferent(oldSize);           // 500
  }                                                                    //
                                                                       //
  var promise2 = promise.then(function (result) {                      // 503
    return parseInt(result.size);                                      // 503
  });                                                                  //
  promise2.cancel = function () {                                      // 504
    promise.cancel();                                                  // 504
  };                                                                   //
                                                                       //
  return promise2;                                                     // 506
};                                                                     //
                                                                       //
Meteor.startup(function () {                                           // 509
  var shutdownApp = function (appId) {                                 // 510
    Grains.find({ appId: appId }).forEach(function (grain) {           // 511
      waitPromise(globalBackend.shutdownGrain(grain._id, grain.userId));
    });                                                                //
  };                                                                   //
                                                                       //
  DevPackages.find().observe({                                         // 516
    removed: function (devPackage) {                                   // 517
      shutdownApp(devPackage.appId);                                   // 517
    },                                                                 //
                                                                       //
    changed: function (oldDevPackage, newDevPackage) {                 // 519
      shutdownApp(oldDevPackage.appId);                                // 520
      if (oldDevPackage.appId !== newDevPackage.appId) {               // 521
        shutdownApp(newDevPackage.appId);                              // 522
      }                                                                //
    },                                                                 //
                                                                       //
    added: function (devPackage) {                                     // 526
      shutdownApp(devPackage.appId);                                   // 526
    }                                                                  //
  });                                                                  //
                                                                       //
  Sessions.find().observe({                                            // 529
    removed: function (session) {                                      // 530
      var proxy = proxiesByHostId[session.hostId];                     // 531
      delete proxiesByHostId[session.hostId];                          // 532
      if (proxy) {                                                     // 533
        proxy.close();                                                 // 534
      }                                                                //
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
// Kill off sessions idle for >~3 minutes.                             //
var TIMEOUT_MS = 180000;                                               // 541
var gcSessions = function () {                                         // 542
  var now = new Date().getTime();                                      // 543
  Sessions.remove({ timestamp: { $lt: now - TIMEOUT_MS } });           // 544
};                                                                     //
                                                                       //
SandstormDb.periodicCleanup(TIMEOUT_MS, gcSessions);                   // 547
                                                                       //
var getProxyForHostId = function (hostId, isAlreadyOpened) {           // 549
  // Get the Proxy corresponding to the given grain session host, possibly (re)creating it if it
  // doesn't already exist. The first request on the session host will always create a new proxy.
  // Later requests may create a proxy if they go to a different front-end replica or if the
  // front-end was restarted.                                          //
  check(hostId, String);                                               // 554
                                                                       //
  return Promise.resolve(undefined).then(function () {                 // 556
    var proxy = proxiesByHostId[hostId];                               // 557
    if (proxy) {                                                       // 558
      return proxy;                                                    // 559
    } else {                                                           //
      // Set table entry to null for now so that we can detect if it is concurrently deleted.
      proxiesByHostId[hostId] = null;                                  // 562
                                                                       //
      return inMeteor(function () {                                    // 564
        var session = Sessions.findOne({ hostId: hostId });            // 565
        if (!session) {                                                // 566
          if (isAlreadyOpened) {                                       // 567
            return new Promise(function (resolve, reject) {            // 568
              var observer = undefined;                                // 569
              var task = Meteor.setTimeout(function () {               // 570
                observer.stop();                                       // 571
                reject(new Meteor.Error(504, "Requested session that no longer exists, and " + "timed out waiting for client to restore it. This can happen if you have " + "opened an app's content in a new window and then closed it in the " + "UI. If you see this error *inside* the Sandstorm UI, please report a " + "bug and describe the circumstances of the error."));
              }, SESSION_PROXY_TIMEOUT);                               //
              observer = Sessions.find({ hostId: hostId }).observe({   // 578
                added: function () {                                   // 579
                  observer.stop();                                     // 580
                  Meteor.clearTimeout(task);                           // 581
                  resolve(getProxyForHostId(hostId, false));           // 582
                }                                                      //
              });                                                      //
            });                                                        //
          } else {                                                     //
            // Does not appear to be a valid session host.             //
            return undefined;                                          // 588
          }                                                            //
        }                                                              //
                                                                       //
        var apiToken = undefined;                                      // 592
        if (session.hashedToken) {                                     // 593
          apiToken = ApiTokens.findOne({ _id: session.hashedToken });  // 594
          // We don't have to fully validate the API token here because if it changed the session
          // would have been deleted.                                  //
          if (!apiToken) {                                             // 597
            throw new Meteor.Error(410, "ApiToken has been deleted");  // 598
          }                                                            //
        }                                                              //
                                                                       //
        var grain = Grains.findOne(session.grainId);                   // 602
        if (!grain) {                                                  // 603
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, "Resource has been deleted");    // 605
        }                                                              //
                                                                       //
        // Note that we don't need to call mayOpenGrain() because the existence of a session
        // implies this check was already performed.                   //
                                                                       //
        var proxy = new Proxy(grain._id, grain.userId, session._id, hostId, session.identityId, false);
        if (apiToken) proxy.apiToken = apiToken;                       // 612
                                                                       //
        // Only add the proxy to the table if it was not concurrently deleted (which could happen
        // e.g. if the user's access was revoked).                     //
        if (hostId in proxiesByHostId) {                               // 616
          proxiesByHostId[hostId] = proxy;                             // 617
        } else {                                                       //
          throw new Meteor.Error(403, "Session was concurrently closed.");
        }                                                              //
                                                                       //
        return proxy;                                                  // 622
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// API tokens                                                          //
                                                                       //
var proxiesByApiToken = {};                                            // 631
                                                                       //
Meteor.startup(function () {                                           // 633
  var clearSessionsAndProxies = function (token) {                     // 634
    // Clears all sessions and API proxies associated with `token` or any token that is downstream
    // in the sharing graph.                                           //
    // TODO(soon): Only clear sessions and proxies for which the permissions have changed.
    var downstream = SandstormPermissions.downstreamTokens(globalDb, { token: token });
    downstream.push(token);                                            // 639
    var identityIds = [];                                              // 640
    var tokenIds = [];                                                 // 641
                                                                       //
    downstream.forEach(function (token) {                              // 643
      var proxy = proxiesByApiToken[token._id];                        // 644
      if (proxy) {                                                     // 645
        proxy.close();                                                 // 646
      }                                                                //
                                                                       //
      delete proxiesByApiToken[token._id];                             // 649
      tokenIds.push(token._id);                                        // 650
      if (token.owner && token.owner.user) {                           // 651
        identityIds.push(token.owner.user.identityId);                 // 652
      }                                                                //
    });                                                                //
                                                                       //
    Sessions.find({                                                    // 656
      grainId: token.grainId,                                          // 657
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    }, {                                                               //
      fields: { hostId: 1 }                                            // 662
    }).forEach(function (session) {                                    //
      var proxy = proxiesByHostId[session.hostId];                     // 664
      if (proxy) {                                                     // 665
        proxy.close();                                                 // 666
      }                                                                //
                                                                       //
      delete proxiesByHostId[session.hostId];                          // 669
    });                                                                //
                                                                       //
    Sessions.remove({                                                  // 672
      grainId: token.grainId,                                          // 673
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    });                                                                //
  };                                                                   //
                                                                       //
  Grains.find().observe({                                              // 681
    changed: function (newGrain, oldGrain) {                           // 682
      if (oldGrain["private"] != newGrain["private"]) {                // 683
        Sessions.remove({ grainId: oldGrain._id, identityId: { $ne: oldGrain.identityId } });
        ApiTokens.find({ grainId: oldGrain._id }).forEach(function (apiToken) {
          delete proxiesByApiToken[apiToken._id];                      // 686
        });                                                            //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  ApiTokens.find({ grainId: { $exists: true }, objectId: { $exists: false } }).observe({
    added: function (newApiToken) {                                    // 693
      // TODO(soon): Unfortunately, added() gets called for all existing role assignments when the
      //   front-end restarts, meaning clearing sessions here will cause people's views to refresh
      //   on server upgrade, which is not a nice user experience. It's also sad to force-refresh
      //   people when they gained new permissions since they might be in the middle of something,
      //   and it's not strictly necessary for security. OTOH, it's sad to be non-reactive. Maybe
      //   we should notify people that they have new permissions and let them click a thing to
      //   refresh?                                                    //
      //clearSessions(roleAssignment.grainId, roleAssignment.recipient);
      //clearApiProxies(roleAssignment.grainId);                       //
    },                                                                 //
                                                                       //
    changed: function (newApiToken, oldApiToken) {                     // 705
      if (!_.isEqual(newApiToken.roleAssignment, oldApiToken.roleAssignment) || !_.isEqual(newApiToken.revoked, oldApiToken.revoked)) {
        clearSessionsAndProxies(newApiToken);                          // 708
      }                                                                //
    },                                                                 //
                                                                       //
    removed: function (oldApiToken) {                                  // 712
      clearSessionsAndProxies(oldApiToken);                            // 713
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
// Used by server/drivers/external-ui/view.js                          //
getProxyForApiToken = function (token) {                               // 719
  check(token, String);                                                // 720
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
  return Promise.resolve(undefined).then(function () {                 // 722
    var proxy = proxiesByApiToken[hashedToken];                        // 723
    if (proxy) {                                                       // 724
      if (proxy.expires && proxy.expires.getTime() <= Date.now()) {    // 725
        throw new Meteor.Error(403, "Authorization token expired");    // 726
      }                                                                //
                                                                       //
      return proxy;                                                    // 729
    } else {                                                           //
      // Set table entry to null for now so that we can detect if it is concurrently deleted.
      proxiesByApiToken[hashedToken] = null;                           // 732
                                                                       //
      return inMeteor(function () {                                    // 734
        var tokenInfo = ApiTokens.findOne(hashedToken);                // 735
        validateWebkey(tokenInfo);                                     // 736
                                                                       //
        var grain = Grains.findOne(tokenInfo.grainId);                 // 738
        if (!grain) {                                                  // 739
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, "Resource has been deleted");    // 741
        }                                                              //
                                                                       //
        var proxy = undefined;                                         // 744
        if (tokenInfo.userInfo) {                                      // 745
          throw new Error("API tokens created with arbitrary userInfo no longer supported");
        } else {                                                       //
          var identityId = null;                                       // 748
          if (tokenInfo.identityId && !tokenInfo.forSharing) {         // 749
            identityId = tokenInfo.identityId;                         // 750
          }                                                            //
                                                                       //
          proxy = new Proxy(tokenInfo.grainId, grain.userId, null, null, identityId, true);
          proxy.apiToken = tokenInfo;                                  // 754
        }                                                              //
                                                                       //
        if (!SandstormPermissions.mayOpenGrain(globalDb, { token: tokenInfo })) {
          // Note that only public grains may be opened without a user ID.
          throw new Meteor.Error(403, "Unauthorized.");                // 759
        }                                                              //
                                                                       //
        if (tokenInfo.expires) {                                       // 762
          proxy.expires = tokenInfo.expires;                           // 763
        }                                                              //
                                                                       //
        // Only add the proxy to the table if it was not concurrently deleted (which could happen
        // e.g. if the token was revoked).                             //
        if (hashedToken in proxiesByApiToken) {                        // 768
          proxiesByApiToken[hashedToken] = proxy;                      // 769
        } else {                                                       //
          throw new Meteor.Error(403, "Token was concurrently revoked.");
        }                                                              //
                                                                       //
        return proxy;                                                  // 774
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
var apiUseBasicAuth = function (req, hostId) {                         // 780
  // If the request was to a token-specific host *and* the request has no Origin header (meaning
  // it is not an XHR from a web site), then we permit the use of HTTP basic auth. The reason we
  // prohibit XHRs is because we want to discourage people from using basic auth in a browser,
  // because the browser will cache the credentials, which exposes the hostname to XSRF attack
  // (which is mostly, but not completely, mitigated by the unguessable hostname). The reason we
  // want to allow basic auth at all is because a lot of existing client apps support only basic
  // auth. New clients and web-based clients should use `Authorization: bearer <token>` instead.
  if (globalDb.isTokenSpecificHostId(hostId) && !req.headers.origin) {
    return true;                                                       // 789
  }                                                                    //
                                                                       //
  // Historically, all API tokens were served from the same host rather than have token-specific
  // hosts. In this model, basic auth in browsers was far more dangerous since the hostname was
  // not a secret -- in fact, even benign attempts to use two APIs on the same server from the same
  // browser could interfere if using basic auth. Hence, basic auth was prohibited except when
  // coming from certain whitelisted clients known not to be browsers, e.g. Mirall, the ownCloud
  // client app.                                                       //
  //                                                                   //
  // Since many clients in the wild have already been configured to use the shared API host, we
  // must continue to support them, so this logic remains.             //
  var agent = req.headers["user-agent"];                               // 801
  return agent.match(BASIC_AUTH_USER_AGENTS_REGEX);                    // 802
};                                                                     //
                                                                       //
var apiTokenForRequest = function (req, hostId) {                      // 805
  // Extract the API token from the request.                           //
                                                                       //
  var auth = req.headers.authorization;                                // 808
  var token = undefined;                                               // 809
  if (auth && auth.slice(0, 7).toLowerCase() === "bearer ") {          // 810
    token = auth.slice(7).trim();                                      // 811
  } else if (auth && auth.slice(0, 6).toLowerCase() === "basic " && apiUseBasicAuth(req, hostId)) {
    token = new Buffer(auth.slice(6).trim(), "base64").toString().split(":")[1];
  } else {                                                             //
    token = undefined;                                                 // 816
  }                                                                    //
                                                                       //
  if (token && hostId !== "api") {                                     // 819
    // Verify that the token matches the specific host.                //
    if (hostId !== globalDb.apiHostIdForToken(token)) {                // 821
      token = undefined;                                               // 822
    }                                                                  //
  }                                                                    //
                                                                       //
  return token;                                                        // 826
};                                                                     //
                                                                       //
// =======================================================================================
// Routing to proxies.                                                 //
//                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyUpgrade = function (hostId, req, socket, head) {               // 834
  // Attempt to handle a WebSocket upgrade by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  if (globalDb.isApiHostId(hostId)) {                                  // 840
    var token = apiTokenForRequest(req, hostId);                       // 841
    if (token) {                                                       // 842
      return getProxyForApiToken(token).then(function (proxy) {        // 843
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 846
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 848
        return true;                                                   // 849
      });                                                              //
    } else {                                                           //
      return Promise.resolve(false);                                   // 852
    }                                                                  //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 857
        // Cross-origin requests are not allowed on UI session hosts.  //
        var origin = req.headers.origin;                               // 859
        if (origin !== PROTOCOL + "//" + req.headers.host) {           // 860
          throw new Meteor.Error(403, "Detected illegal cross-origin WebSocket from: " + origin);
        }                                                              //
                                                                       //
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 866
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 868
        return true;                                                   // 869
      } else {                                                         //
        return false;                                                  // 871
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyRequest = function (hostId, req, res) {                        // 878
  // Attempt to handle an HTTP request by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  if (globalDb.isApiHostId(hostId)) {                                  // 884
    var _ret = (function () {                                          //
      // This is a request for the API host.                           //
                                                                       //
      res.setHeader("Access-Control-Allow-Origin", "*");               // 887
                                                                       //
      if (req.method === "OPTIONS") {                                  // 889
        // Reply to CORS preflight request.                            //
                                                                       //
        // All we want to do is permit APIs to be accessed from arbitrary origins. Since clients
        // must send a valid Authorization header, and since cookies are not used for
        // authorization, this is perfectly safe. In a sane world, we would only need to send back
        // 'Access-Control-Allow-Origin: *' and be done with it.       //
        //                                                             //
        // However, CORS demands that we explicitly whitelist individual methods and headers for
        // use cross-origin, as if this is somehow useful for implementing any practical security
        // policy (it isn't). To make matters worse, we are REQUIRED to enumerate each one
        // individually. We cannot just write '*' for these lists. WTF, CORS?
        //                                                             //
        // Luckily, the request tells us exactly what method and headers are being requested, so we
        // only need to copy those over, rather than create an exhaustive list. But this is still
        // overly complicated.                                         //
                                                                       //
        var accessControlHeaders = {                                   // 906
          "Access-Control-Allow-Methods": "GET, HEAD, POST, PUT, PATCH, DELETE",
          "Access-Control-Max-Age": "3600"                             // 908
        };                                                             //
                                                                       //
        // Copy all requested headers to the allowed headers list.     //
        var requestedHeaders = req.headers["access-control-request-headers"];
        if (requestedHeaders) {                                        // 913
          accessControlHeaders["Access-Control-Allow-Headers"] = requestedHeaders;
        }                                                              //
                                                                       //
        // Add the requested method to the allowed methods list, if it's not there already.
        var requestedMethod = req.headers["access-control-request-method"];
        if (requestedMethod && !_.contains(["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE"], requestedMethod)) {
          accessControlHeaders["Access-Control-Allow-Methods"] += ", " + requestedMethod;
        }                                                              //
                                                                       //
        for (header in babelHelpers.sanitizeForInObject(accessControlHeaders)) {
          res.setHeader(header, accessControlHeaders[header]);         // 925
        }                                                              //
      }                                                                //
                                                                       //
      var errorHandler = function (err) {                              // 929
        if (err instanceof Meteor.Error) {                             // 930
          console.log("error: " + err);                                // 931
          res.writeHead(err.error, err.reason, { "Content-Type": "text/plain" });
        } else {                                                       //
          res.writeHead(500, "Internal Server Error", { "Content-Type": "text/plain" });
        }                                                              //
                                                                       //
        res.end(err.stack);                                            // 937
      };                                                               //
                                                                       //
      var token = apiTokenForRequest(req, hostId);                     // 940
      if (token && req.headers["x-sandstorm-token-keepalive"]) {       // 941
        inMeteor(function () {                                         // 942
          var keepaliveDuration = parseInt(req.headers["x-sandstorm-token-keepalive"]);
          check(keepaliveDuration, Match.Integer);                     // 944
          var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
          validateWebkey(ApiTokens.findOne(hashedToken), new Date(Date.now() + keepaliveDuration));
        }).then(function () {                                          //
          res.writeHead(200, {});                                      // 948
          res.end();                                                   // 949
        }, errorHandler);                                              //
      } else if (token) {                                              //
        getProxyForApiToken(token).then(function (proxy) {             // 952
          proxy.requestHandler(req, res);                              // 953
        }, errorHandler);                                              //
      } else if (req.method === "OPTIONS") {                           //
        // OPTIONS request with no authorization token. Fine to end here.
        res.writeHead(200, {});                                        // 957
        res.end();                                                     // 958
      } else {                                                         //
        if (apiUseBasicAuth(req, hostId)) {                            // 960
          res.writeHead(401, {                                         // 961
            "Content-Type": "text/plain",                              // 962
            "WWW-Authenticate": "Basic realm='Sandstorm API'"          // 963
          });                                                          //
        } else {                                                       //
          // TODO(someday): Display some sort of nifty API browser.    //
          res.writeHead(403, { "Content-Type": "text/plain" });        // 967
        }                                                              //
                                                                       //
        res.end("Missing or invalid authorization header.\n\n" + "This address serves APIs, which allow external apps (such as a phone app) to\n" + "access data on your Sandstorm server. This address is not meant to be opened\n" + "in a regular browser.");
      }                                                                //
                                                                       //
      return {                                                         // 976
        v: Promise.resolve(true)                                       //
      };                                                               //
    })();                                                              //
                                                                       //
    if (typeof _ret === "object") return _ret.v;                       //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 980
        proxy.requestHandler(req, res);                                // 981
        return true;                                                   // 982
      } else {                                                         //
        return false;                                                  // 984
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// =======================================================================================
// Proxy class                                                         //
//                                                                     //
// Connects to a grain and exports it on a wildcard host.              //
//                                                                     //
                                                                       //
var Proxy = (function () {                                             //
  function Proxy(grainId, ownerId, sessionId, hostId, identityId, isApi, supervisor) {
    var _this2 = this;                                                 //
                                                                       //
    babelHelpers.classCallCheck(this, Proxy);                          //
                                                                       //
    this.grainId = grainId;                                            // 998
    this.ownerId = ownerId;                                            // 999
    this.identityId = identityId;                                      // 1000
    this.supervisor = supervisor; // note: optional parameter; we can reconnect
    this.sessionId = sessionId;                                        // 1002
    this.isApi = isApi;                                                // 1003
    this.hasLoaded = false;                                            // 1004
    this.websockets = [];                                              // 1005
    if (sessionId) {                                                   // 1006
      if (!hostId) throw new Error("sessionId must come with hostId");
      if (isApi) throw new Error("API proxy shouldn't have sessionId");
      this.hostId = hostId;                                            // 1009
    } else {                                                           //
      if (!isApi) throw new Error("non-API proxy requires sessionId");
      if (hostId) throw new Error("API proxy sholudn't have hostId");  // 1012
    }                                                                  //
                                                                       //
    if (this.identityId) {                                             // 1015
      var identity = globalDb.getIdentity(this.identityId);            // 1016
      if (!identity) {                                                 // 1017
        throw new Error("identity not found: " + this.identityId);     // 1018
      }                                                                //
                                                                       //
      this.userInfo = {                                                // 1021
        displayName: { defaultText: identity.profile.name },           // 1022
        preferredHandle: identity.profile.handle,                      // 1023
        identityId: new Buffer(identity._id, "hex")                    // 1024
      };                                                               //
      if (identity.profile.pictureUrl) this.userInfo.pictureUrl = identity.profile.pictureUrl;
      if (identity.profile.pronoun) this.userInfo.pronouns = identity.profile.pronoun;
    } else {                                                           //
      this.userInfo = {                                                // 1029
        displayName: { defaultText: "Anonymous User" },                // 1030
        preferredHandle: "anonymous"                                   // 1031
      };                                                               //
    }                                                                  //
                                                                       //
    var _this = this;                                                  // 1035
                                                                       //
    this.requestHandler = function (request, response) {               // 1037
      if (_this2.sessionId) {                                          // 1038
        // Implement /_sandstorm-init for setting the session cookie.  //
        var url = Url.parse(request.url, true);                        // 1040
        if (url.pathname === "/_sandstorm-init" && url.query.sessionid === _this.sessionId) {
          _this.doSessionInit(request, response, url.query.path);      // 1042
          return;                                                      // 1043
        }                                                              //
      }                                                                //
                                                                       //
      Promise.resolve(undefined).then(function () {                    // 1047
        var contentLength = request.headers["content-length"];         // 1048
        if ((request.method === "POST" || request.method === "PUT") && (contentLength === undefined || contentLength > 1024 * 1024)) {
          // The input is either very long, or we don't know how long it is, so use streaming mode.
          return _this.handleRequestStreaming(request, response, contentLength, 0);
        } else {                                                       //
          return readAll(request).then(function (data) {               // 1054
            return _this.handleRequest(request, data, response, 0);    // 1055
          });                                                          //
        }                                                              //
      })["catch"](function (err) {                                     //
        _this.setHasLoaded();                                          // 1059
                                                                       //
        var body = err.stack;                                          // 1061
        if (err.cppFile) {                                             // 1062
          body += "\nC++ location:" + err.cppFile + ":" + (err.line || "??");
        }                                                              //
                                                                       //
        if (err.kjType) {                                              // 1066
          body += "\ntype: " + err.kjType;                             // 1067
        }                                                              //
                                                                       //
        if (response.headersSent) {                                    // 1070
          // Unfortunately, it's too late to tell the client what happened.
          console.error("HTTP request failed after response already sent:", body);
          response.end();                                              // 1073
        } else {                                                       //
          if (err instanceof Meteor.Error) {                           // 1075
            response.writeHead(err.error, err.reason, { "Content-Type": "text/plain" });
          } else {                                                     //
            response.writeHead(500, "Internal Server Error", { "Content-Type": "text/plain" });
          }                                                            //
                                                                       //
          response.end(body);                                          // 1081
        }                                                              //
      });                                                              //
    };                                                                 //
                                                                       //
    this.upgradeHandler = function (request, socket, head) {           // 1086
      _this.handleWebSocket(request, socket, head, 0)["catch"](function (err) {
        console.error("WebSocket setup failed:", err.stack);           // 1088
        // TODO(cleanup):  Manually send back a 500 response?          //
        socket.destroy();                                              // 1090
      });                                                              //
    };                                                                 //
  }                                                                    //
                                                                       //
  Proxy.prototype.close = (function () {                               // 996
    function close() {                                                 // 1095
      this.websockets.forEach(function (socket) {                      // 1096
        socket.destroy();                                              // 1097
      });                                                              //
                                                                       //
      this.websockets = [];                                            // 1100
                                                                       //
      if (this.session) {                                              // 1102
        this.session.close();                                          // 1103
        delete this.session;                                           // 1104
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1107
        this.uiView.close();                                           // 1108
        delete this.uiView;                                            // 1109
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1112
        this.supervisor.close();                                       // 1113
        delete this.supervisor;                                        // 1114
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  Proxy.prototype.getConnection = (function () {                       // 996
    function getConnection() {                                         // 1118
      if (!this.supervisor) {                                          // 1119
        this.supervisor = globalBackend.cap().getGrain(this.ownerId, this.grainId).supervisor;
        this.uiView = null;                                            // 1121
      }                                                                //
                                                                       //
      if (!this.uiView) {                                              // 1124
        this.uiView = this.supervisor.getMainView().view;              // 1125
      }                                                                //
    }                                                                  //
                                                                       //
    return getConnection;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewWebSession = (function () {                  // 996
    function _callNewWebSession(request, userInfo) {                   // 1129
      var params = Capnp.serialize(WebSession.Params, {                // 1130
        basePath: PROTOCOL + "//" + request.headers.host,              // 1131
        userAgent: "user-agent" in request.headers ? request.headers["user-agent"] : "UnknownAgent/0.0",
        acceptableLanguages: "accept-language" in request.headers ? request.headers["accept-language"].split(",").map(function (s) {
          return s.trim();                                             // 1136
        }) : ["en-US", "en"]                                           //
      });                                                              //
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId), WebSession.typeId, params).session;
    }                                                                  //
                                                                       //
    return _callNewWebSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewApiSession = (function () {                  // 996
    function _callNewApiSession(request, userInfo) {                   // 1144
      var _this = this;                                                // 1145
      var params = {};                                                 // 1146
                                                                       //
      if ("x-sandstorm-passthrough" in request.headers) {              // 1148
        var optIns = request.headers["x-sandstorm-passthrough"].split(",").map(function (s) {
          return s.trim();                                             // 1151
        });                                                            //
        // The only currently supported passthrough value is 'address', but others could be useful in
        // the future                                                  //
        if (optIns.indexOf("address") !== -1) {                        // 1154
          // Sadly, we can't use request.socket.remoteFamily because it's not available in the (rather-old)
          // version of node that comes in the Meteor bundle we're using.  Hence this hackery.
          var addressToPass = request.socket.remoteAddress;            // 1157
          if (isRfc1918OrLocal(addressToPass) && "x-real-ip" in request.headers) {
            // Allow overriding the socket's remote address with X-Real-IP header if the request comes
            // from either localhost or an RFC1918 address.  These are not useful for geolocation
            // anyway.                                                 //
            addressToPass = request.headers["x-real-ip"];              // 1162
          }                                                            //
                                                                       //
          if (Net.isIPv4(addressToPass)) {                             // 1165
            // Map IPv4 addresses in IPv6.                             //
            // This conveniently comes out to a 48-bit number, which is precisely representable in a
            // double (which has 53 mantissa bits). Thus we can avoid using Bignum/strings, which we
            // might otherwise need to precisely represent 64-bit fields.
            var v4Int = 0xFFFF00000000 + addressToPass.split(".").map(function (x) {
              return parseInt(x, 10);                                  // 1171
            }).reduce(function (a, b) {                                //
              return 256 * a + b;                                      // 1172
            });                                                        //
            params.remoteAddress = {                                   // 1173
              lower64: v4Int,                                          // 1174
              upper64: 0                                               // 1175
            };                                                         //
          } else if (Net.isIPv6(addressToPass)) {                      //
            // TODO(test): Unit test this                              //
            // Parse a valid v6 address.                               //
            // Split into groups, then insert an appropriate number of 0's if :: was used.
            var groups = addressToPass.split(":");                     // 1181
                                                                       //
            // Strip extra empty group in the case of a leading or trailing '::'.
            if (groups[0] === "") {                                    // 1184
              groups.shift();                                          // 1185
            }                                                          //
                                                                       //
            if (groups[groups.length - 1] === "") {                    // 1188
              groups.pop();                                            // 1189
            }                                                          //
                                                                       //
            var lastGroup = groups[groups.length - 1];                 // 1192
            // Handle IPv4-mapped IPv6 addresses.  These end in a dotted-quad IPv4 address, which we
            // should expand into two groups of 4-character hex strings, like the rest of the address.
            if (Net.isIPv4(lastGroup)) {                               // 1195
              groups.pop();                                            // 1196
              var _quad = lastGroup.split(".").map(function (x) {      // 1197
                return parseInt(x, 10);                                // 1197
              });                                                      //
              groups.push((_quad[0] * 256 + _quad[1]).toString(16));   // 1198
              groups.push((_quad[2] * 256 + _quad[3]).toString(16));   // 1199
            }                                                          //
                                                                       //
            var groupsToAdd = 8 - groups.length;                       // 1202
            var emptyGroupIndex = groups.indexOf("");                  // 1203
            var cleanGroups = undefined;                               // 1204
            if (emptyGroupIndex !== -1) {                              // 1205
              var head = groups.slice(0, emptyGroupIndex);             // 1206
              // groupsToAdd + 1 because we sliced out the empty element
              var mid = Array(groupsToAdd + 1);                        // 1208
              for (var i = 0; i < groupsToAdd + 1; i++) {              // 1209
                mid[i] = "0";                                          // 1210
              }                                                        //
                                                                       //
              var tail = groups.slice(emptyGroupIndex + 1, groups.length);
              cleanGroups = [].concat(head, mid, tail);                // 1214
            } else {                                                   //
              cleanGroups = groups;                                    // 1216
            }                                                          //
                                                                       //
            var ints = cleanGroups.map(function (x) {                  // 1219
              return parseInt(x, 16);                                  // 1219
            });                                                        //
            // We use strings because we'd lose data from loss of precision casting the 64-bit uints
            // into 53-bit-mantissa doubles.                           //
            params.remoteAddress = {                                   // 1222
              upper64: quadToIntString(ints.slice(0, 4)),              // 1223
              lower64: quadToIntString(ints.slice(4, 8))               // 1224
            };                                                         //
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      var serializedParams = Capnp.serialize(ApiSession.Params, params);
                                                                       //
      // TODO(someday): We are currently falling back to WebSession if we get any kind of error upon
      // calling newSession with an ApiSession._id.                    //
      // Eventually we'll remove this logic once we're sure apps have updated.
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId), ApiSession.typeId, serializedParams).then(function (session) {
        return session.session;                                        // 1239
      }, function (err) {                                              //
        return _this._callNewWebSession(request, userInfo);            // 1241
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewApiSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewSession = (function () {                     // 996
    function _callNewSession(request, viewInfo) {                      // 1246
      var userInfo = _.clone(this.userInfo);                           // 1247
      var _this = this;                                                // 1248
      var promise = inMeteor(function () {                             // 1249
        var vertex = undefined;                                        // 1250
        if (_this.apiToken) {                                          // 1251
          vertex = { token: _this.apiToken };                          // 1252
        } else {                                                       //
          // (_this.identityId might be null; this is fine)            //
          vertex = { grain: { _id: _this.grainId, identityId: _this.identityId } };
        }                                                              //
                                                                       //
        var permissions = SandstormPermissions.grainPermissions(globalDb, vertex, viewInfo);
        if (!permissions) {                                            // 1259
          throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");
        }                                                              //
                                                                       //
        Sessions.update({                                              // 1263
          _id: _this.sessionId                                         // 1264
        }, {                                                           //
          $set: {                                                      // 1266
            viewInfo: viewInfo,                                        // 1267
            permissions: permissions                                   // 1268
          }                                                            //
        });                                                            //
                                                                       //
        return permissions;                                            // 1272
      });                                                              //
                                                                       //
      return promise.then(function (permissions) {                     // 1275
        userInfo.permissions = permissions;                            // 1276
                                                                       //
        var numBytes = Math.ceil(permissions.length / 8);              // 1278
        var buf = new Buffer(numBytes);                                // 1279
        for (var ii = 0; ii < numBytes; ++ii) {                        // 1280
          buf[ii] = 0;                                                 // 1281
        }                                                              //
                                                                       //
        for (var ii = 0; ii < permissions.length; ++ii) {              // 1284
          var byteNum = Math.floor(ii / 8);                            // 1285
          var bitNum = ii % 8;                                         // 1286
          if (permissions[ii]) {                                       // 1287
            buf[byteNum] = buf[byteNum] | 1 << bitNum;                 // 1288
          }                                                            //
        }                                                              //
                                                                       //
        userInfo.deprecatedPermissionsBlob = buf;                      // 1292
                                                                       //
        if (_this.isApi) {                                             // 1294
          return _this._callNewApiSession(request, userInfo);          // 1295
        } else {                                                       //
          return _this._callNewWebSession(request, userInfo);          // 1297
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewSession;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.getSession = (function () {                          // 996
    function getSession(request) {                                     // 1302
      var _this3 = this;                                               //
                                                                       //
      if (!this.session) {                                             // 1303
        (function () {                                                 //
          _this3.getConnection(); // make sure we're connected         // 1304
          var _this = _this3;                                          // 1305
          var promise = _this3.uiView.getViewInfo().then(function (viewInfo) {
            return inMeteor(function () {                              // 1307
              Grains.update(_this.grainId, { $set: { cachedViewInfo: viewInfo } });
            }).then(function () {                                      //
              return _this._callNewSession(request, viewInfo);         // 1310
            });                                                        //
          }, function (error) {                                        //
            if (error.kjType === "failed" || error.kjType === "unimplemented") {
              // Method not implemented.                               //
              // TODO(apibump): Don't treat 'failed' as 'unimplemented'. Unfortunately, old apps built
              //   with old versions of Cap'n Proto don't throw 'unimplemented' exceptions, so we have
              //   to accept 'failed' here at least until the next API bump.
              return _this._callNewSession(request, {});               // 1318
            } else {                                                   //
              return Promise.reject(error);                            // 1320
            }                                                          //
          });                                                          //
          _this3.session = new Capnp.Capability(promise, WebSession);  // 1323
        })();                                                          //
      }                                                                //
                                                                       //
      return this.session;                                             // 1326
    }                                                                  //
                                                                       //
    return getSession;                                                 //
  })();                                                                //
                                                                       //
  Proxy.prototype.keepAlive = (function () {                           // 996
    function keepAlive() {                                             // 1329
      this.getConnection();                                            // 1330
      return this.supervisor.keepAlive();                              // 1331
    }                                                                  //
                                                                       //
    return keepAlive;                                                  //
  })();                                                                //
                                                                       //
  Proxy.prototype.resetConnection = (function () {                     // 996
    function resetConnection() {                                       // 1334
      if (this.session) {                                              // 1335
        this.session.close();                                          // 1336
        delete this.session;                                           // 1337
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1340
        this.uiView.close();                                           // 1341
        delete this.uiView;                                            // 1342
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1345
        this.supervisor.close();                                       // 1346
        delete this.supervisor;                                        // 1347
      }                                                                //
    }                                                                  //
                                                                       //
    return resetConnection;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.maybeRetryAfterError = (function () {                // 996
    function maybeRetryAfterError(error, retryCount) {                 // 1351
      // If the error may be caused by the grain dying or a network failure, try to restart it,
      // returning a promise that resolves once restarted. Otherwise, just rethrow the error.
      // `retryCount` should be incremented for every successful retry as part of the same request;
      // we only want to retry once.                                   //
      var _this = this;                                                // 1356
      if (SandstormBackend.shouldRestartGrain(error, retryCount)) {    // 1357
        this.resetConnection();                                        // 1358
        return inMeteor(function () {                                  // 1359
          _this.supervisor = globalBackend.continueGrain(_this.grainId).supervisor;
        });                                                            //
      } else {                                                         //
        throw error;                                                   // 1363
      }                                                                //
    }                                                                  //
                                                                       //
    return maybeRetryAfterError;                                       //
  })();                                                                //
                                                                       //
  Proxy.prototype.doSessionInit = (function () {                       // 996
    function doSessionInit(request, response, requestPath) {           // 1367
      // jscs:disable disallowQuotedKeysInObjects                      //
      var path = requestPath || "/";                                   // 1369
                                                                       //
      // Check that the path is relative (ie. starts with a /).        //
      // Also ensure that it doesn't start with 2 /, because that is interpreted as non-relative
      if (path.lastIndexOf("/", 0) !== 0 || path.lastIndexOf("//", 0) === 0) {
        response.writeHead(400, "Invalid path supplied", { "Content-Type": "text/plain" });
        response.end("Invalid path supplied.");                        // 1375
        return;                                                        // 1376
      }                                                                //
                                                                       //
      // Set the session ID.                                           //
      response.setHeader("Set-Cookie", ["sandstorm-sid=", this.sessionId, "; Max-Age=31536000; HttpOnly"].join(""));
                                                                       //
      response.setHeader("Cache-Control", "no-cache, private");        // 1382
                                                                       //
      // Redirect to the app's root URL.                               //
      // Note:  All browsers support relative locations and the next update to HTTP/1.1 will officially
      //   make them valid.  http://tools.ietf.org/html/draft-ietf-httpbis-p2-semantics-26#page-67
      response.writeHead(303, "See Other", { "Location": encodeURI(path) });
      response.end();                                                  // 1388
    }                                                                  //
                                                                       //
    return doSessionInit;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.makeContext = (function () {                         // 996
    function makeContext(request, response) {                          // 1391
      // Parses the cookies from the request, checks that the session ID is present and valid, then
      // returns the request context which contains the other cookies.  Throws an exception if the
      // session ID is missing or invalid.                             //
                                                                       //
      var context = {};                                                // 1396
                                                                       //
      if (this.hostId) {                                               // 1398
        var parseResult = parseCookies(request);                       // 1399
        if (!parseResult.sessionId || parseResult.sessionId !== this.sessionId) {
          throw new Meteor.Error(403, "Unauthorized");                 // 1401
        }                                                              //
                                                                       //
        if (parseResult.cookies.length > 0) {                          // 1404
          context.cookies = parseResult.cookies;                       // 1405
        }                                                              //
      } else {                                                         //
        // jscs:disable disallowEmptyBlocks                            //
        // This is an API request. Cookies are not supported.          //
      }                                                                //
                                                                       //
      context.accept = parseAcceptHeader(request);                     // 1412
                                                                       //
      context.eTagPrecondition = parsePreconditionHeader(request);     // 1414
                                                                       //
      context.additionalHeaders = [];                                  // 1416
      WebSession.Context.headerWhitelist.forEach(function (headerName) {
        if (request.headers[headerName]) {                             // 1418
          context.additionalHeaders.push({                             // 1419
            name: headerName,                                          // 1420
            value: request.headers[headerName]                         // 1421
          });                                                          //
        };                                                             //
      });                                                              //
                                                                       //
      var promise = new Promise(function (resolve, reject) {           // 1426
        response.resolveResponseStream = resolve;                      // 1427
        response.rejectResponseStream = reject;                        // 1428
      });                                                              //
                                                                       //
      context.responseStream = new Capnp.Capability(promise, ByteStream);
                                                                       //
      return context;                                                  // 1433
    }                                                                  //
                                                                       //
    return makeContext;                                                //
  })();                                                                //
                                                                       //
  Proxy.prototype.translateResponse = (function () {                   // 996
    function translateResponse(rpcResponse, response, request) {       // 1436
      if (this.hostId) {                                               // 1437
        if (rpcResponse.setCookies && rpcResponse.setCookies.length > 0) {
          response.setHeader("Set-Cookie", rpcResponse.setCookies.map(makeSetCookieHeader));
        }                                                              //
                                                                       //
        // TODO(security): Add a Content-Security-Policy header which:
        // (1) Prevents the app from initiating HTTP requests to third parties.
        // (2) Prevents the app from navigating the parent frame.      //
        // (3) Prevents the app from opening popups.                   //
        // (4) Prohibits anyone other than the Sandstorm shell from framing the app (as a backup
        //   defense vs. clickjacking, though unguessable hostnames already mostly prevent this).
      } else {                                                         //
          // jscs:disable validateQuoteMarks                           //
          // This is an API request. Cookies are not supported.        //
                                                                       //
          // We need to make sure caches know that different bearer tokens get totally different results.
          response.setHeader('Vary', 'Authorization');                 // 1453
                                                                       //
          // APIs can be called from any origin. Because we ignore cookies, there is no security problem.
          response.setHeader('Access-Control-Allow-Origin', '*');      // 1456
                                                                       //
          // Add a Content-Security-Policy as a backup in case someone finds a way to load this resource
          // in a browser context. This policy should thoroughly neuter it.
          response.setHeader('Content-Security-Policy', "default-src 'none'; sandbox");
        }                                                              //
                                                                       //
      // On first response, update the session to have hasLoaded=true  //
      this.setHasLoaded();                                             // 1464
                                                                       //
      // TODO(security): Set X-Content-Type-Options: nosniff?          //
                                                                       //
      if ('content' in rpcResponse) {                                  // 1468
        var content = rpcResponse.content;                             // 1469
        var code = successCodes[content.statusCode];                   // 1470
        if (!code) {                                                   // 1471
          throw new Error('Unknown status code: ', content.statusCode);
        }                                                              //
                                                                       //
        if (content.mimeType) {                                        // 1475
          response.setHeader('Content-Type', content.mimeType);        // 1476
        }                                                              //
                                                                       //
        if (content.encoding) {                                        // 1479
          response.setHeader('Content-Encoding', content.encoding);    // 1480
        }                                                              //
                                                                       //
        if (content.language) {                                        // 1483
          response.setHeader('Content-Language', content.language);    // 1484
        }                                                              //
                                                                       //
        if (content.eTag) {                                            // 1487
          response.setHeader('ETag', composeETag(content.eTag));       // 1488
        }                                                              //
                                                                       //
        if ('disposition' in content && 'download' in content.disposition) {
          response.setHeader('Content-Disposition', 'attachment; filename="' + content.disposition.download.replace(/([\\"\n])/g, "\\$1") + '"');
        }                                                              //
                                                                       //
        if ('stream' in content.body) {                                // 1496
          if (request.method === 'HEAD') {                             // 1497
            content.body.stream.close();                               // 1498
            response.rejectResponseStream(new Error('HEAD request; content doesn\'t matter.'));
          } else {                                                     //
            var _ret3 = (function () {                                 //
              var streamHandle = content.body.stream;                  // 1501
              response.writeHead(code.id, code.title);                 // 1502
              var promise = new Promise(function (resolve, reject) {   // 1503
                response.resolveResponseStream(new Capnp.Capability(new ResponseStream(response, streamHandle, resolve, reject), ByteStream));
              });                                                      //
              promise.streamHandle = streamHandle;                     // 1507
              return {                                                 // 1508
                v: promise                                             //
              };                                                       //
            })();                                                      //
                                                                       //
            if (typeof _ret3 === "object") return _ret3.v;             //
          }                                                            //
        } else {                                                       //
          response.rejectResponseStream(new Error('Response content body was not a stream.'));
                                                                       //
          if ('bytes' in content.body) {                               // 1514
            response.setHeader('Content-Length', content.body.bytes.length);
          } else {                                                     //
            throw new Error('Unknown content body type.');             // 1517
          }                                                            //
        }                                                              //
                                                                       //
        response.writeHead(code.id, code.title);                       // 1521
                                                                       //
        if ('bytes' in content.body && request.method !== 'HEAD') {    // 1523
          response.write(content.body.bytes);                          // 1524
        }                                                              //
                                                                       //
        response.end();                                                // 1527
      } else if ('noContent' in rpcResponse) {                         //
        var noContent = rpcResponse.noContent;                         // 1529
        var noContentCode = noContentSuccessCodes[noContent.shouldResetForm * 1];
        response.writeHead(noContentCode.id, noContentCode.title);     // 1531
        response.end();                                                // 1532
      } else if ('preconditionFailed' in rpcResponse) {                //
        var preconditionFailed = rpcResponse.preconditionFailed;       // 1534
        if (request.method === 'GET' && 'if-none-match' in request.headers) {
          if (preconditionFailed.matchingETag) {                       // 1536
            response.setHeader('ETag', composeETag(preconditionFailed.matchingETag));
          }                                                            //
                                                                       //
          response.writeHead(304, 'Not Modified');                     // 1540
        } else {                                                       //
          response.writeHead(412, 'Precondition Failed');              // 1542
        }                                                              //
                                                                       //
        response.end();                                                // 1545
      } else if ('redirect' in rpcResponse) {                          //
        var redirect = rpcResponse.redirect;                           // 1547
        var redirectCode = redirectCodes[redirect.switchToGet * 2 + redirect.isPermanent];
        response.writeHead(redirectCode.id, redirectCode.title, {      // 1549
          'Location': redirect.location                                // 1550
        });                                                            //
        response.end();                                                // 1552
      } else if ('clientError' in rpcResponse) {                       //
        var clientError = rpcResponse.clientError;                     // 1554
        var errorCode = errorCodes[clientError.statusCode];            // 1555
        if (!errorCode) {                                              // 1556
          throw new Error('Unknown status code: ', clientError.statusCode);
        }                                                              //
                                                                       //
        response.writeHead(errorCode.id, errorCode.title, {            // 1560
          'Content-Type': 'text/html'                                  // 1561
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1564
          if (clientError.descriptionHtml) {                           // 1565
            response.write(clientError.descriptionHtml);               // 1566
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>' + errorCode.id + ': ' + errorCode.title + '</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1574
      } else if ('serverError' in rpcResponse) {                       //
        response.writeHead(500, 'Internal Server Error', {             // 1576
          'Content-Type': 'text/html'                                  // 1577
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1580
          if (rpcResponse.serverError.descriptionHtml) {               // 1581
            response.write(rpcResponse.serverError.descriptionHtml);   // 1582
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>500: Internal Server Error</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1589
      } else {                                                         //
        throw new Error('Unknown HTTP response type:\n' + JSON.stringify(rpcResponse));
      }                                                                //
                                                                       //
      return Promise.resolve(undefined);                               // 1594
    }                                                                  //
                                                                       //
    return translateResponse;                                          //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequest = (function () {                       // 996
    function handleRequest(request, data, response, retryCount) {      // 1597
      var _this = this;                                                // 1598
      return Promise.resolve(undefined).then(function () {             // 1599
        return _this.makeContext(request, response);                   // 1600
      }).then(function (context) {                                     //
        // jscs:disable requireDotNotation                             //
        // Send the RPC.                                               //
        var path = request.url.slice(1); // remove leading '/'         // 1604
        var session = _this.getSession(request);                       // 1605
                                                                       //
        var requestContent = function () {                             // 1607
          return {                                                     // 1608
            content: data,                                             // 1609
            encoding: request.headers['content-encoding'],             // 1610
            mimeType: request.headers['content-type']                  // 1611
          };                                                           //
        };                                                             //
                                                                       //
        var xmlContent = function () {                                 // 1615
          var type = request.headers['content-type'] || 'application/xml;charset=utf-8';
          var match = type.match(/[^/]*\/xml(; *charset *= *([^ ;]*))?/);
          if (!match) {                                                // 1618
            response.writeHead(415, 'Unsupported media type.', {       // 1619
              'Content-Type': 'text/plain'                             // 1620
            });                                                        //
            response.end('expected XML request body');                 // 1622
            throw new Error('expected XML request body');              // 1623
          }                                                            //
                                                                       //
          var charset = match[2] || 'ISO-8859-1';                      // 1626
                                                                       //
          var encoding = request.headers['content-encoding'];          // 1628
          if (encoding && encoding !== 'identity') {                   // 1629
            if (encoding !== 'gzip') throw new Error('unknown Content-Encoding: ' + encoding);
            data = gunzipSync(data);                                   // 1631
          }                                                            //
                                                                       //
          return data.toString(charset.toLowerCase() === 'utf-8' ? 'utf8' : 'binary');
        };                                                             //
                                                                       //
        var propfindDepth = function () {                              // 1637
          var depth = request.headers['depth'];                        // 1638
          return depth === '0' ? 'zero' : depth === '1' ? 'one' : 'infinity';
        };                                                             //
                                                                       //
        var shallow = function () {                                    // 1644
          return request.headers['depth'] === '0';                     // 1645
        };                                                             //
                                                                       //
        var noOverwrite = function () {                                // 1648
          return (request.headers['overwrite'] || '').toLowerCase() === 'f';
        };                                                             //
                                                                       //
        var destination = function () {                                // 1652
          var result = request.headers['destination'];                 // 1653
          if (!result) throw new Error('missing destination');         // 1654
          return Url.parse(result).path.slice(1); // remove leading '/'
        };                                                             //
                                                                       //
        if (request.method === 'GET' || request.method === 'HEAD') {   // 1658
          return session.get(path, context, request.method === 'HEAD');
        } else if (request.method === 'POST') {                        //
          return session.post(path, requestContent(), context);        // 1661
        } else if (request.method === 'PUT') {                         //
          return session.put(path, requestContent(), context);         // 1663
        } else if (request.method === 'PATCH') {                       //
          return session.patch(path, requestContent(), context);       // 1665
        } else if (request.method === 'DELETE') {                      //
          return session["delete"](path, context);                     // 1667
        } else if (request.method === 'PROPFIND') {                    //
          return session.propfind(path, xmlContent(), propfindDepth(), context);
        } else if (request.method === 'PROPPATCH') {                   //
          return session.proppatch(path, xmlContent(), context);       // 1671
        } else if (request.method === 'MKCOL') {                       //
          return session.mkcol(path, requestContent(), context);       // 1673
        } else if (request.method === 'COPY') {                        //
          return session.copy(path, destination(), noOverwrite(), shallow(), context);
        } else if (request.method === 'MOVE') {                        //
          return session.move(path, destination(), noOverwrite(), context);
        } else if (request.method === 'LOCK') {                        //
          return session.lock(path, xmlContent(), shallow(), context);
        } else if (request.method === 'UNLOCK') {                      //
          return session.unlock(path, request.headers['lock-token'], context);
        } else if (request.method === 'ACL') {                         //
          return session.acl(path, xmlContent(), context);             // 1683
        } else if (request.method === 'REPORT') {                      //
          return session.report(path, requestContent(), context);      // 1685
        } else if (request.method === 'OPTIONS') {                     //
          return session.options(path, context).then(function (options) {
            var dav = [];                                              // 1688
            if (options.davClass1) dav.push('1');                      // 1689
            if (options.davClass2) dav.push('2');                      // 1690
            if (options.davClass3) dav.push('3');                      // 1691
            if (dav.length > 0) response.setHeader('DAV', dav.join(', '));
            response.end();                                            // 1693
            // Return no response; we already handled everything.      //
          }, function (err) {                                          //
            if (err.kjType !== 'unimplemented') throw err;             // 1696
            response.end();                                            // 1697
            // Return no response; we already handled everything.      //
          });                                                          //
        } else {                                                       //
            throw new Error('Sandstorm only supports the following methods: GET, POST, PUT, PATCH, DELETE, HEAD, PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK, ACL, REPORT, and OPTIONS.');
          }                                                            //
      }).then(function (rpcResponse) {                                 //
        if (rpcResponse !== undefined) {                               // 1704
          // Will be undefined for OPTIONS request.                    //
          return _this.translateResponse(rpcResponse, response, request);
        }                                                              //
      })["catch"](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleRequest(request, data, response, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequest;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequestStreaming = (function () {              // 996
    function handleRequestStreaming(request, response, contentLength, retryCount) {
      var _this = this;                                                // 1715
      var context = this.makeContext(request, response);               // 1716
      var path = request.url.slice(1); // remove leading '/'           // 1717
      var session = this.getSession(request);                          // 1718
                                                                       //
      var mimeType = request.headers['content-type'] || 'application/octet-stream';
      var encoding = request.headers['content-encoding'];              // 1721
                                                                       //
      var requestStreamPromise = undefined;                            // 1723
      if (request.method === 'POST') {                                 // 1724
        requestStreamPromise = session.postStreaming(path, mimeType, context, encoding);
      } else if (request.method === 'PUT') {                           //
        requestStreamPromise = session.putStreaming(path, mimeType, context, encoding);
      } else {                                                         //
        throw new Error('Sandstorm only supports streaming POST and PUT requests.');
      }                                                                //
                                                                       //
      // TODO(perf): We ought to be pipelining the body, but we can't currently, because we have to
      //   handle the case where the app doesn't actually support streaming. We could pipeline while
      //   also buffering the data on the side in case we need it again later, but that's kind of
      //   complicated. We should fix the whole protocol to make streaming the standard.
      return requestStreamPromise.then(function (requestStreamResult) {
        var requestStream = requestStreamResult.stream;                // 1737
                                                                       //
        // Initialized when getResponse() returns, if the response is streaming.
        var downloadStreamHandle = undefined;                          // 1740
                                                                       //
        // Initialized if an upload-stream method throws.              //
        var uploadStreamError = undefined;                             // 1743
                                                                       //
        // We call `getResponse()` immediately so that the app can start streaming data down even while
        // data is still being streamed up. This theoretically allows apps to perform bidirectional
        // streaming, though probably very few actually do that.       //
        //                                                             //
        // Note that we need to be able to cancel `responsePromise` below, so it's important that it is
        // the raw Cap'n Proto promise. Hence `translateResponsePromise` is a separate variable.
        var responsePromise = requestStream.getResponse();             // 1751
                                                                       //
        var reportUploadStreamError = function (err) {                 // 1753
          // Called when an upload-stream method throws.               //
                                                                       //
          if (!uploadStreamError) {                                    // 1756
            uploadStreamError = err;                                   // 1757
                                                                       //
            // If we're still waiting on any response stuff, cancel it.
            responsePromise.cancel();                                  // 1760
            requestStream.close();                                     // 1761
            if (downloadStreamHandle) {                                // 1762
              downloadStreamHandle.close();                            // 1763
            }                                                          //
          }                                                            //
        };                                                             //
                                                                       //
        // If we have a Content-Length, pass it along to the app by calling `expectSize()`.
        if (contentLength !== undefined) {                             // 1769
          requestStream.expectSize(contentLength)["catch"](function (err) {
            // expectSize() is allowed to be unimplemented.            //
            if (err.kjType !== 'unimplemented') {                      // 1772
              reportUploadStreamError(err);                            // 1773
            }                                                          //
          });                                                          //
        }                                                              //
                                                                       //
        // Pipe the input stream to the app.                           //
        request.on('data', function (buf) {                            // 1779
          // TODO(soon): Only allow a small number of write()s to be in-flight at once,
          //   pausing the input stream if we hit that limit, so that we block the TCP socket all the
          //   way back to the source. May want to also coalesce small writes for this purpose.
          // TODO(security): The above problem may allow a DoS attack on the front-end.
          if (!uploadStreamError) requestStream.write(buf)["catch"](reportUploadStreamError);
        });                                                            //
                                                                       //
        request.on('end', function () {                                // 1787
          if (!uploadStreamError) requestStream.done()["catch"](reportUploadStreamError);
                                                                       //
          // We're all done making calls to requestStream.             //
          requestStream.close();                                       // 1791
        });                                                            //
                                                                       //
        request.on('close', function () {                              // 1794
          reportUploadStreamError(new Error('HTTP connection unexpectedly closed during request.'));
        });                                                            //
                                                                       //
        request.on('error', function (err) {                           // 1798
          reportUploadStreamError(err);                                // 1799
        });                                                            //
                                                                       //
        return responsePromise.then(function (rpcResponse) {           // 1802
          // Stop here if the upload stream has already failed.        //
          if (uploadStreamError) throw uploadStreamError;              // 1804
          var promise = _this.translateResponse(rpcResponse, response, request);
          downloadStreamHandle = promise.streamHandle;                 // 1806
          return promise;                                              // 1807
        });                                                            //
      }, function (err) {                                              //
        if (err.kjType === 'failed' && err.message.indexOf('not implemented') !== -1) {
          // Hack to work around old apps using an old version of Cap'n Proto, before the
          // 'unimplemented' exception type was introduced. :(         //
          // TODO(cleanup): When we transition to API version 2, we can move this into the
          //   compatibility layer.                                    //
          err.kjType = 'unimplemented';                                // 1815
        }                                                              //
                                                                       //
        if (SandstormBackend.shouldRestartGrain(err, 0)) {             // 1818
          // This is the kind of error that indicates we should retry. Note that we passed 0 for the
          // retry count above because we were just checking if this is a retriable error (vs. possibly
          // a method-not-implemented error); maybeRetryAfterError() will check again with the proper
          // retry count.                                              //
          return _this.maybeRetryAfterError(err, retryCount).then(function () {
            return _this.handleRequestStreaming(request, response, contentLength, retryCount + 1);
          });                                                          //
        } else if (err.kjType === 'unimplemented') {                   //
          // Streaming is not implemented. Fall back to non-streaming version.
          return readAll(request).then(function (data) {               // 1828
            return _this.handleRequest(request, data, response, 0);    // 1829
          });                                                          //
        } else {                                                       //
          throw err;                                                   // 1832
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequestStreaming;                                     //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleWebSocket = (function () {                     // 996
    function handleWebSocket(request, socket, head, retryCount) {      // 1837
      var _this = this;                                                // 1838
                                                                       //
      return Promise.resolve(undefined).then(function () {             // 1840
        return _this.makeContext(request);                             // 1841
      }).then(function (context) {                                     //
        var path = request.url.slice(1); // remove leading '/'         // 1843
        var session = _this.getSession(request);                       // 1844
                                                                       //
        if (!('sec-websocket-key' in request.headers)) {               // 1846
          throw new Error('Missing Sec-WebSocket-Accept header.');     // 1847
        }                                                              //
                                                                       //
        var magic = request.headers['sec-websocket-key'] + '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
        var acceptKey = Crypto.createHash('sha1').update(magic).digest('base64');
                                                                       //
        var protocols = [];                                            // 1853
        if ('sec-websocket-protocol' in request.headers) {             // 1854
          protocols = request.headers['sec-websocket-protocol'].split(',').map(function (s) {
            return s.trim();                                           // 1856
          });                                                          //
        }                                                              //
                                                                       //
        var receiver = new WebSocketReceiver(socket);                  // 1859
        // TODO(someday): do we want to make these be weak references somehow?
        _this.websockets.push(socket);                                 // 1861
                                                                       //
        var promise = session.openWebSocket(path, context, protocols, receiver);
                                                                       //
        if (head.length > 0) {                                         // 1865
          promise.serverStream.sendBytes(head);                        // 1866
        }                                                              //
                                                                       //
        pumpWebSocket(socket, promise.serverStream);                   // 1869
                                                                       //
        return promise.then(function (response) {                      // 1871
          var headers = ['HTTP/1.1 101 Switching Protocols', 'Upgrade: websocket', 'Connection: Upgrade', 'Sec-WebSocket-Accept: ' + acceptKey];
          if (response.protocol && response.protocol.length > 0) {     // 1878
            headers.push('Sec-WebSocket-Protocol: ' + response.protocol.join(', '));
          }                                                            //
                                                                       //
          headers.push('');                                            // 1882
          headers.push('');                                            // 1883
                                                                       //
          socket.write(headers.join('\r\n'));                          // 1885
          receiver.go();                                               // 1886
                                                                       //
          // Note:  At this point errors are out of our hands.         //
        });                                                            //
      })["catch"](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleWebSocket(request, socket, head, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleWebSocket;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.setHasLoaded = (function () {                        // 996
    function setHasLoaded() {                                          // 1897
      var _this4 = this;                                               //
                                                                       //
      if (!this.hasLoaded) {                                           // 1898
        (function () {                                                 //
          _this4.hasLoaded = true;                                     // 1899
          var sessionId = _this4.sessionId;                            // 1900
          inMeteor(function () {                                       // 1901
            Sessions.update({ _id: sessionId }, { $set: { hasLoaded: true } });
          });                                                          //
        })();                                                          //
      }                                                                //
    }                                                                  //
                                                                       //
    return setHasLoaded;                                               //
  })();                                                                //
                                                                       //
  return Proxy;                                                        //
})();                                                                  //
                                                                       //
;                                                                      // 1906
                                                                       //
var PROTOCOL = Url.parse(process.env.ROOT_URL).protocol;               // 1908
                                                                       //
var isRfc1918OrLocal = function (address) {                            // 1910
  if (Net.isIPv4(address)) {                                           // 1911
    quad = address.split('.').map(function (x) {                       // 1912
      return parseInt(x, 10);                                          // 1912
    });                                                                //
    return quad[0] === 127 || quad[0] === 10 || quad[0] === 192 && quad[1] === 168 || quad[0] === 172 && quad[1] >= 16 && quad[1] < 32;
  } else if (Net.isIPv6(address)) {                                    //
    // IPv6 specifies ::1 as localhost and fd:: as reserved for private networks
    return address === '::1' || address.lastIndexOf('fd', 0) === 0;    // 1918
  } else {                                                             //
    // Ignore things that are neither IPv4 nor IPv6                    //
    return false;                                                      // 1921
  }                                                                    //
};                                                                     //
                                                                       //
var quadToIntString = function (quad) {                                // 1925
  var num = Bignum(quad[0]).shiftLeft(16).add(quad[1]).shiftLeft(16).add(quad[2]).shiftLeft(16).add(quad[3]);
  return num.toString();                                               // 1930
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Session cookie management                                           //
                                                                       //
var parseCookies = function (request) {                                // 1936
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['cookie'];                              // 1938
                                                                       //
  var result = { cookies: [] };                                        // 1940
  if (header) {                                                        // 1941
    var reqCookies = header.split(';');                                // 1942
    for (var i in babelHelpers.sanitizeForInObject(reqCookies)) {      // 1943
      var reqCookie = reqCookies[i];                                   // 1944
      var equalsPos = reqCookie.indexOf('=');                          // 1945
      var cookie = undefined;                                          // 1946
      if (equalsPos === -1) {                                          // 1947
        cookie = { key: reqCookie.trim(), value: '' };                 // 1948
      } else {                                                         //
        cookie = { key: reqCookie.slice(0, equalsPos).trim(), value: reqCookie.slice(equalsPos + 1) };
      }                                                                //
                                                                       //
      if (cookie.key === 'sandstorm-sid') {                            // 1953
        if (result.sessionId) {                                        // 1954
          throw new Error('Multiple sandstorm session IDs?');          // 1955
        }                                                              //
                                                                       //
        result.sessionId = cookie.value;                               // 1958
      } else {                                                         //
        result.cookies.push(cookie);                                   // 1960
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 1965
};                                                                     //
                                                                       //
var parsePreconditionHeader = function (request) {                     // 1968
  if (request.headers['if-match']) {                                   // 1969
    if (request.headers['if-match'].trim() === '*') {                  // 1970
      return { exists: null };                                         // 1971
    }                                                                  //
                                                                       //
    var matches = parseETagList(request.headers['if-match']);          // 1974
    if (matches.length > 0) {                                          // 1975
      return { matchesOneOf: matches };                                // 1976
    }                                                                  //
  }                                                                    //
                                                                       //
  if (request.headers['if-none-match']) {                              // 1980
    if (request.headers['if-none-match'].trim() === '*') {             // 1981
      return { doesntExist: null };                                    // 1982
    }                                                                  //
                                                                       //
    var noneMatches = parseETagList(request.headers['if-none-match']);
    if (noneMatches.length > 0) {                                      // 1986
      return { matchesNoneOf: noneMatches };                           // 1987
    }                                                                  //
  }                                                                    //
                                                                       //
  return { none: null };                                               // 1991
};                                                                     //
                                                                       //
var parseETagList = function (input) {                                 // 1994
  // An ETag is a quoted, \-escaped string, possibly prefixed with W/ (outside the quotes) to
  // indicate that it is weak. We are parsing a list of comma-delimited etags.
                                                                       //
  input = input.trim();                                                // 1998
  var results = [];                                                    // 1999
                                                                       //
  while (input.length > 0) {                                           // 2001
    var match = input.match(/^\s*(W\/)?"(([^"\\]|\\.)*)"\s*($|,)/);    // 2002
    if (!match) throw new Meteor.Error(400, 'invalid etag');           // 2003
                                                                       //
    input = input.slice(match[0].length).trim();                       // 2005
    results.push({ weak: !!match[1], value: match[2].replace(/\\(.)/g, '$1') });
  }                                                                    //
                                                                       //
  return results;                                                      // 2009
};                                                                     //
                                                                       //
var composeETag = function (tag) {                                     // 2012
  var result = '"' + (tag.value || '').replace(/([\\"])/g, '\\$1') + '"';
  if (tag.weak) result = 'W/' + result;                                // 2014
  return result;                                                       // 2015
};                                                                     //
                                                                       //
var parseAcceptHeader = function (request) {                           // 2018
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['accept'];                              // 2020
                                                                       //
  var result = [];                                                     // 2022
  if (header) {                                                        // 2023
    var acceptList = header.split(',');                                // 2024
    for (var i in babelHelpers.sanitizeForInObject(acceptList)) {      // 2025
      var acceptStr = acceptList[i];                                   // 2026
      var tokensList = acceptStr.split(';');                           // 2027
                                                                       //
      var temp = { mimeType: tokensList[0].trim() };                   // 2029
                                                                       //
      var tokensListRest = tokensList.slice(1);                        // 2031
      for (var j in babelHelpers.sanitizeForInObject(tokensListRest)) {
        var token = tokensListRest[j];                                 // 2033
        var equalsPos = token.indexOf('=');                            // 2034
        if (equalsPos) {                                               // 2035
          var key = token.slice(0, equalsPos).trim();                  // 2036
          var value = token.slice(equalsPos + 1).trim();               // 2037
                                                                       //
          if (key === 'q') {                                           // 2039
            temp.qValue = +value;                                      // 2040
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      result.push(temp);                                               // 2045
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 2049
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Regular HTTP request handling                                       //
                                                                       //
var readAll = function (stream) {                                      // 2055
  return new Promise(function (resolve, reject) {                      // 2056
    var buffers = [];                                                  // 2057
    stream.on('data', function (buf) {                                 // 2058
      buffers.push(buf);                                               // 2059
    });                                                                //
                                                                       //
    stream.on('end', function () {                                     // 2062
      resolve(Buffer.concat(buffers));                                 // 2063
    });                                                                //
                                                                       //
    stream.on('error', reject);                                        // 2066
  });                                                                  //
};                                                                     //
                                                                       //
var makeSetCookieHeader = function (cookie) {                          // 2070
  var result = [cookie.name, '=', cookie.value];                       // 2071
                                                                       //
  if ('absolute' in cookie.expires) {                                  // 2073
    result.push('; Expires=');                                         // 2074
    result.push(new Date(cookie.expires.absolute * 1000).toUTCString());
  } else if ('relative' in cookie.expires) {                           //
    result.push('; Max-Age=' + cookie.expires.relative);               // 2077
  }                                                                    //
                                                                       //
  if (cookie.path) {                                                   // 2080
    result.push('; Path=' + cookie.path);                              // 2081
  }                                                                    //
                                                                       //
  if (cookie.httpOnly) {                                               // 2084
    result.push('; HttpOnly');                                         // 2085
  }                                                                    //
                                                                       //
  return result.join('');                                              // 2088
};                                                                     //
                                                                       //
// TODO(cleanup):  Auto-generate based on annotations in web-session.capnp.
var successCodes = {                                                   // 2092
  ok: { id: 200, title: 'OK' },                                        // 2093
  created: { id: 201, title: 'Created' },                              // 2094
  accepted: { id: 202, title: 'Accepted' },                            // 2095
  multiStatus: { id: 207, title: 'Multi-Status' }                      // 2096
};                                                                     //
var noContentSuccessCodes = [                                          // 2098
// Indexed by shouldResetForm * 1                                      //
{ id: 204, title: 'No Content' }, { id: 205, title: 'Reset Content' }];
var redirectCodes = [                                                  // 2103
// Indexed by switchToGet * 2 + isPermanent                            //
{ id: 307, title: 'Temporary Redirect' }, { id: 308, title: 'Permanent Redirect' }, { id: 303, title: 'See Other' }, { id: 301, title: 'Moved Permanently' }];
var errorCodes = {                                                     // 2110
  badRequest: { id: 400, title: 'Bad Request' },                       // 2111
  forbidden: { id: 403, title: 'Forbidden' },                          // 2112
  notFound: { id: 404, title: 'Not Found' },                           // 2113
  methodNotAllowed: { id: 405, title: 'Method Not Allowed' },          // 2114
  notAcceptable: { id: 406, title: 'Not Acceptable' },                 // 2115
  conflict: { id: 409, title: 'Conflict' },                            // 2116
  gone: { id: 410, title: 'Gone' },                                    // 2117
  requestEntityTooLarge: { id: 413, title: 'Request Entity Too Large' },
  requestUriTooLong: { id: 414, title: 'Request-URI Too Long' },       // 2119
  unsupportedMediaType: { id: 415, title: 'Unsupported Media Type' },  // 2120
  imATeapot: { id: 418, title: 'I\'m a teapot' },                      // 2121
  unprocessableEntity: { id: 422, title: 'Unprocessable Entity' }      // 2122
};                                                                     //
                                                                       //
ResponseStream = (function () {                                        // 2125
  function ResponseStream(response, streamHandle, resolve, reject) {   // 2126
    babelHelpers.classCallCheck(this, ResponseStream);                 //
                                                                       //
    this.response = response;                                          // 2127
    this.streamHandle = streamHandle;                                  // 2128
    this.resolve = resolve;                                            // 2129
    this.reject = reject;                                              // 2130
    this.ended = false;                                                // 2131
  }                                                                    //
                                                                       //
  ResponseStream.prototype.write = (function () {                      // 2125
    function write(data) {                                             // 2134
      this.response.write(data);                                       // 2135
    }                                                                  //
                                                                       //
    return write;                                                      //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.done = (function () {                       // 2125
    function done() {                                                  // 2138
      this.response.end();                                             // 2139
      this.streamHandle.close();                                       // 2140
      this.ended = true;                                               // 2141
    }                                                                  //
                                                                       //
    return done;                                                       //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.close = (function () {                      // 2125
    function close() {                                                 // 2144
      if (this.ended) {                                                // 2145
        this.resolve();                                                // 2146
      } else {                                                         //
        this.streamHandle.close();                                     // 2148
        this.reject(new Error('done() was never called on outbound stream.'));
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return ResponseStream;                                               //
})();                                                                  //
                                                                       //
// TODO(cleanup): Node 0.12 has a `gunzipSync` but 0.10 (which Meteor still uses) does not.
var Zlib = Npm.require('zlib');                                        // 2155
var gunzipSync = Meteor.wrapAsync(Zlib.gunzip, Zlib);                  // 2156
                                                                       //
// -----------------------------------------------------------------------------
// WebSocket handling                                                  //
                                                                       //
WebSocketReceiver = (function () {                                     // 2161
  function WebSocketReceiver(socket) {                                 // 2162
    babelHelpers.classCallCheck(this, WebSocketReceiver);              //
                                                                       //
    this.socket = socket;                                              // 2163
    this.queue = [];                                                   // 2164
  }                                                                    //
                                                                       //
  WebSocketReceiver.prototype.go = (function () {                      // 2161
    function go() {                                                    // 2167
      for (var i in babelHelpers.sanitizeForInObject(this.queue)) {    // 2168
        this.socket.write(this.queue[i]);                              // 2169
      }                                                                //
                                                                       //
      this.queue = null;                                               // 2172
    }                                                                  //
                                                                       //
    return go;                                                         //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.sendBytes = (function () {               // 2161
    function sendBytes(message) {                                      // 2175
      // TODO(someday):  Flow control of some sort?                    //
      if (this.queue === null) {                                       // 2177
        this.socket.write(message);                                    // 2178
      } else {                                                         //
        this.queue.push(message);                                      // 2180
      }                                                                //
    }                                                                  //
                                                                       //
    return sendBytes;                                                  //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.close = (function () {                   // 2161
    function close() {                                                 // 2184
      this.socket.end();                                               // 2185
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return WebSocketReceiver;                                            //
})();                                                                  //
                                                                       //
pumpWebSocket = function (socket, rpcStream) {                         // 2189
  socket.on('data', function (chunk) {                                 // 2190
    rpcStream.sendBytes(chunk)["catch"](function (err) {               // 2191
      if (err.kjType !== 'disconnected') {                             // 2192
        console.error('WebSocket sendBytes failed: ' + err.stack);     // 2193
      }                                                                //
                                                                       //
      socket.destroy();                                                // 2196
    });                                                                //
  });                                                                  //
                                                                       //
  socket.on('end', function (chunk) {                                  // 2200
    rpcStream.close();                                                 // 2201
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// Debug log access                                                    //
                                                                       //
Meteor.publish('grainLog', function (grainId) {                        // 2208
  var _this5 = this;                                                   //
                                                                       //
  check(grainId, String);                                              // 2209
  var id = 0;                                                          // 2210
  var grain = Grains.findOne(grainId);                                 // 2211
  if (!grain || !this.userId || grain.userId !== this.userId) {        // 2212
    this.added('grainLog', id++, { text: 'Only the grain owner can view the debug log.' });
    this.ready();                                                      // 2214
    return;                                                            // 2215
  }                                                                    //
                                                                       //
  var connected = false;                                               // 2218
  var _this = this;                                                    // 2219
                                                                       //
  var receiver = {                                                     // 2221
    write: function (data) {                                           // 2222
      connected = true;                                                // 2223
      _this.added('grainLog', id++, { text: data.toString('utf8') });  // 2224
    },                                                                 //
                                                                       //
    close: function () {                                               // 2227
      if (connected) {                                                 // 2228
        _this.added('grainLog', id++, {                                // 2229
          text: '*** lost connection to grain (probably because it shut down) ***'
        });                                                            //
      }                                                                //
    }                                                                  //
  };                                                                   //
                                                                       //
  try {                                                                // 2236
    (function () {                                                     //
      var handle = waitPromise(globalBackend.useGrain(grainId, function (supervisor) {
        return supervisor.watchLog(8192, receiver);                    // 2238
      })).handle;                                                      //
      connected = true;                                                // 2240
      _this5.onStop(function () {                                      // 2241
        handle.close();                                                // 2242
      });                                                              //
    })();                                                              //
  } catch (err) {                                                      //
    if (!connected) {                                                  // 2245
      this.added('grainLog', id++, {                                   // 2246
        text: '*** couldn\'t connect to grain (' + err + ') ***'       // 2247
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  // Notify ready.                                                     //
  this.ready();                                                        // 2253
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=proxy.js.map
