(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/devAccounts.js                                               //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var allowDevAccounts = Meteor.settings && Meteor.settings["public"] && Meteor.settings["public"].allowDevAccounts;
                                                                       //
Accounts.identityServices.dev = {                                      // 20
  isEnabled: function () {                                             // 21
    return allowDevAccounts;                                           // 22
  },                                                                   //
                                                                       //
  loginTemplate: {                                                     // 25
    name: "devLoginForm",                                              // 26
    priority: -10 }                                                    // 27
};                                                                     //
                                                                       //
// Put it at the top.                                                  //
if (allowDevAccounts) {                                                // 31
  if (Meteor.isServer) {                                               // 32
    var Crypto = Npm.require("crypto");                                // 33
                                                                       //
    Meteor.methods({                                                   // 35
      createDevAccount: function (displayName, isAdmin, profile, unverifiedEmail) {
        // This is a login method that creates or logs in a dev account with the given displayName
                                                                       //
        check(displayName, String);                                    // 39
        check(isAdmin, Match.OneOf(undefined, null, Boolean));         // 40
        check(profile, Match.OneOf(undefined, null, Object));          // 41
        check(unverifiedEmail, Match.OneOf(undefined, null, String));  // 42
                                                                       //
        isAdmin = isAdmin || false;                                    // 44
                                                                       //
        profile = profile || {};                                       // 46
        profile.name = profile.name || displayName;                    // 47
        var hasCompletedSignup = !!unverifiedEmail && !!profile.pronoun && !!profile.handle;
                                                                       //
        var user = Meteor.users.findOne({ "services.dev.name": displayName });
        var userId = undefined;                                        // 51
                                                                       //
        if (user) {                                                    // 53
          userId = user._id;                                           // 54
        } else {                                                       //
          userId = Accounts.insertUserDoc({ profile: profile,          // 56
            unverifiedEmail: unverifiedEmail }, { services: { dev: { name: displayName,
                isAdmin: isAdmin,                                      // 60
                hasCompletedSignup: hasCompletedSignup } } });         // 61
        }                                                              //
        // Log them in on this connection.                             //
        return Accounts._loginMethod(this, "createDevAccount", arguments, "dev", function () {
          return { userId: userId };                                   // 65
        });                                                            //
      }                                                                //
    });                                                                //
  }                                                                    //
                                                                       //
  if (Meteor.isClient) {                                               // 70
    loginDevAccount = function (displayName, isAdmin) {                // 71
      Accounts.callLoginMethod({                                       // 72
        methodName: "createDevAccount",                                // 73
        methodArguments: [displayName, isAdmin],                       // 74
        userCallback: function (err) {                                 // 75
          if (err) {                                                   // 76
            window.alert(err);                                         // 77
          }                                                            //
        }                                                              //
      });                                                              //
    };                                                                 //
                                                                       //
    loginDevAccountFast = function (displayName, isAdmin) {            // 83
      return new Promise(function (resolve, reject) {                  // 84
        // This skips the firstSignUp page. Mostly used for testing purposes.
        var profile = {                                                // 86
          name: displayName,                                           // 87
          pronoun: "robot",                                            // 88
          handle: "_" + displayName.toLowerCase()                      // 89
        };                                                             //
                                                                       //
        Accounts.callLoginMethod({                                     // 92
          methodName: "createDevAccount",                              // 93
          methodArguments: [displayName, isAdmin, profile, displayName + "@example.com"],
          userCallback: function (err) {                               // 95
            if (err) {                                                 // 96
              reject(new Error(err));                                  // 97
            } else {                                                   //
              Router.go("apps");                                       // 99
              resolve();                                               // 100
            }                                                          //
          }                                                            //
        });                                                            //
      });                                                              //
    };                                                                 //
  }                                                                    //
}                                                                      //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=devAccounts.js.map
