(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/grain.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file implements /grain, i.e. the main view into an app.        //
                                                                       //
var DEFAULT_TITLE = "Sandstorm";                                       // 19
                                                                       //
if (Meteor.isServer) {                                                 // 21
  (function () {                                                       //
    var Crypto = Npm.require("crypto");                                // 22
                                                                       //
    Meteor.publish("grainTopBar", function (grainId) {                 // 24
      check(grainId, String);                                          // 25
                                                                       //
      var result = [Grains.find({                                      // 27
        _id: grainId,                                                  // 29
        $or: [{ userId: this.userId }, { "private": { $ne: true } }]   // 30
      }, {                                                             //
        fields: {                                                      // 35
          title: 1,                                                    // 36
          userId: 1,                                                   // 37
          identityId: 1,                                               // 38
          "private": 1                                                 // 39
        }                                                              //
      })];                                                             //
      if (this.userId) {                                               // 43
        var myIdentityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
        result.push(ApiTokens.find({                                   // 45
          grainId: grainId,                                            // 46
          $or: [{ "owner.user.identityId": { $in: myIdentityIds } }, { identityId: { $in: myIdentityIds } }]
        }));                                                           //
      }                                                                //
                                                                       //
      return result;                                                   // 54
    });                                                                //
                                                                       //
    // We allow users to learn package information about a grain they own.
    // This is used for obtaining icon and app title information for grains
    // you own, which is used in the sidebar. It is not a security/privacy
    // risk since it only exposes this information for grains the user owns.
    Meteor.publish("packageByGrainId", function (grainId) {            // 61
      check(grainId, String);                                          // 62
      var publishThis = [];                                            // 63
      // We need to publish the packageId so that client-side code can
      // find the right package.                                       //
      var thisGrainCursor = Grains.find({                              // 66
        _id: grainId,                                                  // 67
        userId: this.userId                                            // 68
      }, {                                                             //
        fields: { packageId: 1 }                                       // 70
      });                                                              //
      publishThis.push(thisGrainCursor);                               // 72
                                                                       //
      if (thisGrainCursor.count()) {                                   // 74
        var thisGrain = thisGrainCursor.fetch()[0];                    // 75
        var thisPackageCursor = Packages.find({ _id: thisGrain.packageId });
        publishThis.push(thisPackageCursor);                           // 77
      }                                                                //
                                                                       //
      return publishThis;                                              // 80
    });                                                                //
                                                                       //
    Meteor.publish("tokenInfo", function (token) {                     // 83
      // Allows the client side to map a raw token to its entry in ApiTokens, and the additional
      // metadata that it will need to display the app icon and title.  We do not care about making
      // the metadata reactive.                                        //
      check(token, String);                                            // 87
                                                                       //
      var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
      var apiToken = ApiTokens.findOne({                               // 90
        _id: hashedToken                                               // 91
      }, {                                                             //
        fields: {                                                      // 93
          grainId: 1,                                                  // 94
          owner: 1                                                     // 95
        }                                                              //
      });                                                              //
      if (!apiToken) {                                                 // 98
        this.added("tokenInfo", token, { invalidToken: true });        // 99
      } else {                                                         //
        var grainId = apiToken.grainId;                                // 101
        var grain = Grains.findOne({                                   // 102
          _id: grainId                                                 // 103
        }, {                                                           //
          fields: {                                                    // 105
            packageId: 1,                                              // 106
            appId: 1                                                   // 107
          }                                                            //
        });                                                            //
        if (!grain) {                                                  // 110
          this.added("tokenInfo", token, { invalidToken: true });      // 111
        } else {                                                       //
          if (apiToken.owner && apiToken.owner.user) {                 // 113
            var identity = globalDb.getIdentity(apiToken.owner.user.identityId);
            var metadata = apiToken.owner.user.denormalizedGrainMetadata;
            if (identity && metadata) {                                // 116
              this.added("tokenInfo", token, { identityOwner: _.pick(identity, "_id", "profile"),
                grainId: grainId,                                      // 119
                grainMetadata: metadata });                            // 120
            } else {                                                   //
              this.added("tokenInfo", token, { invalidToken: true });  // 122
            }                                                          //
          } else if (!apiToken.owner || "webkey" in apiToken.owner) {  //
            var pkg = Packages.findOne({ _id: grain.packageId }, { fields: { manifest: 1 } });
            var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || { defaultText: "" };
            var appIcon = undefined;                                   // 127
            if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {
              var icons = pkg.manifest.metadata.icons;                 // 129
              appIcon = icons.grain || icons.appGrid;                  // 130
            }                                                          //
                                                                       //
            var denormalizedGrainMetadata = {                          // 133
              appTitle: appTitle,                                      // 134
              icon: appIcon,                                           // 135
              appId: appIcon ? undefined : grain.appId                 // 136
            };                                                         //
            this.added("tokenInfo", token, { webkey: true,             // 138
              grainId: grainId,                                        // 140
              grainMetadata: denormalizedGrainMetadata });             // 141
          } else {                                                     //
            this.added("tokenInfo", token, { invalidToken: true });    // 143
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      this.ready();                                                    // 148
      return;                                                          // 149
    });                                                                //
                                                                       //
    Meteor.publish("requestingAccess", function (grainId, identityId) {
      check(grainId, String);                                          // 153
      check(identityId, String);                                       // 154
                                                                       //
      if (!this.userId) {                                              // 156
        throw new Meteor.Error(403, "Must be logged in to request access.");
      }                                                                //
                                                                       //
      if (!globalDb.userHasIdentity(this.userId, identityId)) {        // 160
        throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
      }                                                                //
                                                                       //
      var grain = globalDb.getGrain(grainId);                          // 164
      if (!grain) {                                                    // 165
        throw new Meteor.Error(404, "Grain not found.");               // 166
      }                                                                //
                                                                       //
      var _this = this;                                                // 169
      var query = ApiTokens.find({ grainId: grainId, accountId: grain.userId,
        "owner.user.identityId": identityId,                           // 171
        revoked: { $ne: true } });                                     // 172
      var handle = query.observe({                                     // 173
        added: function (apiToken) {                                   // 174
          _this.added("grantedAccessRequests", Random.id(), { grainId: grainId, identityId: identityId });
        }                                                              //
      });                                                              //
                                                                       //
      this.onStop(function () {                                        // 180
        return handle.stop();                                          //
      });                                                              //
    });                                                                //
                                                                       //
    Meteor.publish("grainSize", function (grainId) {                   // 183
      // Publish pseudo-collection containing the size of the grain opened in the given session.
      check(grainId, String);                                          // 185
                                                                       //
      var grain = Grains.findOne(grainId);                             // 187
      if (!grain || grain.userId !== this.userId) {                    // 188
        return [];                                                     // 189
      }                                                                //
                                                                       //
      var supervisor = globalBackend.cap().getGrain(this.userId, grainId).supervisor;
                                                                       //
      var _this = this;                                                // 194
      var stopped = false;                                             // 195
      var promise = getGrainSize(supervisor);                          // 196
                                                                       //
      function getNext(oldSize) {                                      // 198
        promise = getGrainSize(supervisor, oldSize);                   // 199
        promise.then(function (size) {                                 // 200
          if (!stopped) {                                              // 201
            if (size !== oldSize) {                                    // 202
              // sometimes there are false alarms                      //
              _this.changed("grainSizes", grainId, { size: size });    // 203
            }                                                          //
                                                                       //
            getNext(size);                                             // 206
          }                                                            //
        }, function (err) {                                            //
                                                                       //
          if (!stopped) {                                              // 210
            if (err.kjType === "disconnected") {                       // 211
              _this.stop();                                            // 212
            } else {                                                   //
              _this.error(err);                                        // 214
            }                                                          //
          }                                                            //
        });                                                            //
      }                                                                //
                                                                       //
      promise.then(function (size) {                                   // 220
        if (!stopped) {                                                // 221
          _this.added("grainSizes", grainId, { size: size });          // 222
          _this.ready();                                               // 223
          getNext(size);                                               // 224
        }                                                              //
      }, function (err) {                                              //
                                                                       //
        if (!stopped) {                                                // 228
          if (err.kjType === "disconnected") {                         // 229
            _this.stop();                                              // 230
          } else {                                                     //
            _this.error(err);                                          // 232
          }                                                            //
        }                                                              //
      });                                                              //
                                                                       //
      _this.onStop(function () {                                       // 237
        stopped = true;                                                // 238
        promise.cancel();                                              // 239
      });                                                              //
    });                                                                //
  })();                                                                //
}                                                                      //
                                                                       //
// GrainSizes is used by grainview.js                                  //
GrainSizes = new Mongo.Collection("grainSizes");                       // 245
// TokenInfo is used by grainview.js                                   //
TokenInfo = new Mongo.Collection("tokenInfo");                         // 247
GrantedAccessRequests = new Mongo.Collection("grantedAccessRequests");
// Pseudo-collections published above.                                 //
                                                                       //
function emailLinkWithInlineStyle(url, text) {                         // 251
  return "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + text + "</a>";
}                                                                      //
                                                                       //
Meteor.methods({                                                       // 258
  deleteGrain: function (grainId) {                                    // 259
    check(grainId, String);                                            // 260
                                                                       //
    if (this.userId) {                                                 // 262
      var grain = Grains.findOne({ _id: grainId, userId: this.userId });
      if (grain) {                                                     // 264
        Grains.remove(grainId);                                        // 265
        ApiTokens.remove({                                             // 266
          grainId: grainId,                                            // 267
          $or: [{ owner: { $exists: false } }, { owner: { webkey: null } }]
        });                                                            //
        if (grain.lastUsed) {                                          // 273
          DeleteStats.insert({ type: isDemoUser() ? "demoGrain" : "grain",
            lastActive: grain.lastUsed, appId: grain.appId });         // 275
        }                                                              //
                                                                       //
        if (!this.isSimulation) {                                      // 278
          waitPromise(globalBackend.deleteGrain(grainId, this.userId));
          Meteor.call("deleteUnusedPackages", grain.appId);            // 280
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
                                                                       //
  forgetGrain: function (grainId, identityId) {                        // 286
    // TODO(cleanup): For now we are ignoring `identityId`, but maybe we should expose finer-grained
    //  forgetting.                                                    //
                                                                       //
    check(grainId, String);                                            // 290
    check(identityId, String);                                         // 291
                                                                       //
    if (!this.userId) {                                                // 293
      throw new Meteor.Error(403, "Must be logged in to forget a grain.");
    }                                                                  //
                                                                       //
    SandstormDb.getUserIdentityIds(Meteor.user()).forEach(function (identityId) {
      ApiTokens.remove({ grainId: grainId, "owner.user.identityId": identityId });
    });                                                                //
  },                                                                   //
                                                                       //
  updateGrainPreferredIdentity: function (grainId, identityId) {       // 302
    check(grainId, String);                                            // 303
    check(identityId, String);                                         // 304
    if (!this.userId) {                                                // 305
      throw new Meteor.Error(403, "Must be logged in.");               // 306
    }                                                                  //
                                                                       //
    var grain = globalDb.getGrain(grainId) || {};                      // 309
    if (!grain.userId === this.userId) {                               // 310
      throw new Meteor.Error(403, "Grain not owned by current user.");
    }                                                                  //
                                                                       //
    Grains.update({ _id: grainId }, { $set: { identityId: identityId } });
  },                                                                   //
                                                                       //
  updateGrainTitle: function (grainId, newTitle, identityId) {         // 317
    check(grainId, String);                                            // 318
    check(newTitle, String);                                           // 319
    check(identityId, String);                                         // 320
    if (this.userId) {                                                 // 321
      var grain = Grains.findOne(grainId);                             // 322
      if (grain) {                                                     // 323
        if (grain.userId === this.userId) {                            // 324
          Grains.update({ _id: grainId, userId: this.userId }, { $set: { title: newTitle } });
        } else {                                                       //
          if (!globalDb.userHasIdentity(this.userId, identityId)) {    // 327
            throw new Meteor.Error(403, "Current user does not have identity " + identityId);
          }                                                            //
                                                                       //
          var token = ApiTokens.findOne({                              // 331
            grainId: grainId,                                          // 332
            objectId: { $exists: false },                              // 333
            "owner.user.identityId": identityId                        // 334
          }, {                                                         //
            sort: { created: 1 }                                       // 336
          });                                                          //
          if (token) {                                                 // 338
            ApiTokens.update(token._id, { $set: { "owner.user.title": newTitle } });
          }                                                            //
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
                                                                       //
  privatizeGrain: function (grainId) {                                 // 346
    check(grainId, String);                                            // 347
    if (this.userId) {                                                 // 348
      Grains.update({ _id: grainId, userId: this.userId }, { $set: { "private": true } });
    }                                                                  //
  },                                                                   //
                                                                       //
  inviteUsersToGrain: function (origin, identityId, grainId, title, roleAssignment, contacts, message) {
    var _this2 = this;                                                 //
                                                                       //
    if (!this.isSimulation) {                                          // 355
      var _ret2 = (function () {                                       //
        check(origin, String);                                         // 356
        check(identityId, String);                                     // 357
        check(grainId, String);                                        // 358
        check(title, String);                                          // 359
        check(roleAssignment, roleAssignmentPattern);                  // 360
        check(contacts, [{                                             // 361
          _id: String,                                                 // 363
          isDefault: Match.Optional(Boolean),                          // 364
          profile: Match.ObjectIncluding({                             // 365
            service: String,                                           // 366
            name: String,                                              // 367
            intrinsicName: String                                      // 368
          })                                                           //
        }]);                                                           //
        check(message, { text: String, html: String });                // 372
        if (!_this2.userId) {                                          // 373
          throw new Meteor.Error(403, "Must be logged in to share by email.");
        }                                                              //
                                                                       //
        if (!globalDb.userHasIdentity(_this2.userId, identityId)) {    // 377
          throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
        }                                                              //
                                                                       //
        if (contacts.length === 0) {                                   // 381
          throw new Meteor.Error(400, "No contacts were provided.");   // 382
        }                                                              //
                                                                       //
        var accountId = _this2.userId;                                 // 385
        var identity = globalDb.getIdentity(identityId);               // 386
        var sharerDisplayName = identity.profile.name;                 // 387
        var outerResult = { successes: [], failures: [] };             // 388
        contacts.forEach(function (contact) {                          // 389
          if (contact.isDefault && contact.profile.service === "email") {
            var emailAddress = contact.profile.intrinsicName;          // 391
            var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "email invitation for " + emailAddress, roleAssignment, { webkey: { forSharing: true } });
            var _url = origin + "/shared/" + result.token;             // 396
            var html = message.html + "<br><br>" + emailLinkWithInlineStyle(_url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: If you forward this email to other people, they will be able to access " + "the share as well. To prevent this, remove the button before forwarding.</div>";
            try {                                                      // 402
              SandstormEmail.send({                                    // 403
                to: emailAddress,                                      // 404
                from: "Sandstorm server <no-reply@" + HOSTNAME + ">",  // 405
                subject: sharerDisplayName + " has invited you to join a grain: " + title,
                text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + _url + "\n\nNote: If you forward this email to other people, they will be able to " + "access the share as well. To prevent this, remove the link before forwarding.",
                html: html                                             // 410
              });                                                      //
            } catch (e) {                                              //
              outerResult.failures.push({ contact: contact, error: e.toString() });
            }                                                          //
          } else {                                                     //
            var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "direct invitation to " + contact.profile.intrinsicName, roleAssignment, { user: { identityId: contact._id, title: title } });
            try {                                                      // 420
              var _identity = Meteor.users.findOne({ _id: contact._id });
              var emailAddress = SandstormDb.getVerifiedEmails(_identity)[0];
              var _url2 = origin + "/shared/" + result.token;          // 423
              if (emailAddress) {                                      // 424
                var intrinsicName = contact.profile.intrinsicName;     // 425
                var loginNote = undefined;                             // 426
                if (contact.profile.service === "google") {            // 427
                  loginNote = "Google account with address " + emailAddress;
                } else if (contact.profile.service === "github") {     //
                  loginNote = "Github account with username " + intrinsicName;
                } else if (contact.profile.service === "email") {      //
                  loginNote = "email address " + intrinsicName;        // 432
                } else {                                               //
                  throw new Meteor.Error(500, "Unknown service to email share link.");
                }                                                      //
                                                                       //
                var html = message.html + "<br><br>" + emailLinkWithInlineStyle(_url2, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: You will need to log in with your " + loginNote + " to access this grain.";
                SandstormEmail.send({                                  // 442
                  to: emailAddress,                                    // 443
                  from: "Sandstorm server <no-reply@" + HOSTNAME + ">",
                  subject: sharerDisplayName + " has invited you to join a grain: " + title,
                  text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + _url2 + "\n\nNote: You will need to log in with your " + loginNote + " to access this grain.",
                  html: html                                           // 449
                });                                                    //
              } else {                                                 //
                outerResult.failures.push({ contact: contact, warning: "User does not have a " + "verified email, so notification of this share was not sent to them. Please " + "manually share " + _url2 + " with them." });
              }                                                        //
            } catch (e) {                                              //
              outerResult.failures.push({ contact: contact, error: e.toString(),
                warning: "Share succeeded, but there was an error emailing the user. Please " + "manually share " + url + " with them." });
            }                                                          //
          }                                                            //
        });                                                            //
                                                                       //
        return {                                                       // 464
          v: outerResult                                               //
        };                                                             //
      })();                                                            //
                                                                       //
      if (typeof _ret2 === "object") return _ret2.v;                   //
    }                                                                  //
  },                                                                   //
                                                                       //
  requestAccess: function (origin, grainId, identityId) {              // 468
    check(origin, String);                                             // 469
    check(grainId, String);                                            // 470
    check(identityId, String);                                         // 471
    if (!this.isSimulation) {                                          // 472
      if (!this.userId) {                                              // 473
        throw new Meteor.Error(403, "Must be logged in to request access.");
      }                                                                //
                                                                       //
      if (!globalDb.userHasIdentity(this.userId, identityId)) {        // 477
        throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
      }                                                                //
                                                                       //
      var grain = Grains.findOne(grainId);                             // 481
      if (!grain) {                                                    // 482
        throw new Meteor.Error(404, "No such grain");                  // 483
      }                                                                //
                                                                       //
      var grainOwner = globalDb.getUser(grain.userId);                 // 486
      var email = _.findWhere(SandstormDb.getUserEmails(grainOwner), { primary: true });
      if (!email) {                                                    // 488
        throw new Meteor.Error("no email", "Grain owner has no email address.");
      }                                                                //
                                                                       //
      var emailAddress = email.email;                                  // 492
                                                                       //
      var identity = globalDb.getIdentity(identityId);                 // 494
      globalDb.addContact(grainOwner._id, identityId);                 // 495
                                                                       //
      var envelopeFrom = globalDb.getReturnAddress();                  // 497
      var fromEmail = globalDb.getServerTitle() + " <" + globalDb.getReturnAddress() + ">";
      var senderEmails = SandstormDb.getVerifiedEmails(identity);      // 499
      if (senderEmails.length > 0) {                                   // 500
        var primaryEmail = Meteor.user().primaryEmail;                 // 501
        if (senderEmails.indexOf(primaryEmail) >= 0) {                 // 502
          fromEmail = primaryEmail;                                    // 503
        } else {                                                       //
          fromEmail = senderEmails[0];                                 // 505
        }                                                              //
      }                                                                //
                                                                       //
      // TODO(soon): In the HTML version, we should display an identity card.
      var identityNote = "";                                           // 510
      if (identity.profile.service === "google") {                     // 511
        identityNote = " (" + identity.privateIntrinsicName + ")";     // 512
      } else if (identity.profile.service === "github") {              //
        identityNote = " (" + identity.profile.intrinsicName + " on GitHub)";
      } else if (identity.profile.service === "email") {               //
        identityNote = " (" + identity.profile.intrinsicName + ")";    // 516
      }                                                                //
                                                                       //
      var message = identity.profile.name + identityNote + " is requesting access to your grain: " + grain.title + ".";
                                                                       //
      var _url3 = origin + "/share/" + grainId + "/" + identityId;     // 522
                                                                       //
      var html = message + "<br><br>" + emailLinkWithInlineStyle(_url3, "Open Sharing Menu");
                                                                       //
      var user = Meteor.user();                                        // 527
      var ACCESS_REQUEST_LIMIT = 10;                                   // 528
      var resetCount = true;                                           // 529
      if (user.accessRequests) {                                       // 530
        if (user.accessRequests.resetOn < new Date()) {                // 531
          Meteor.users.update({ _id: user._id }, { $unset: { accessRequests: 1 } });
        } else if (user.accessRequests.count >= ACCESS_REQUEST_LIMIT) {
          throw new Meteor.Error(403, "For spam control reasons, you are not allowed to make " + "more than " + ACCESS_REQUEST_LIMIT + " access requests per day.");
        } else {                                                       //
          resetCount = false;                                          // 538
        }                                                              //
      }                                                                //
                                                                       //
      var modifier = { $inc: { "accessRequests.count": 1 } };          // 542
      if (resetCount) {                                                // 543
        var tomorrow = new Date(Date.now() + 86400000);                // 544
        modifier.$set = { "accessRequests.resetOn": tomorrow };        // 545
      }                                                                //
                                                                       //
      Meteor.users.update({ _id: user._id }, modifier);                // 548
                                                                       //
      SandstormEmail.send({                                            // 550
        to: emailAddress,                                              // 551
        envelopeFrom: envelopeFrom,                                    // 552
        from: fromEmail,                                               // 553
        subject: grain.title + " - Request for access",                // 554
        text: message + "\n\nFollow this link to share access:\n\n" + _url3,
        html: html                                                     // 556
      });                                                              //
    }                                                                  //
  }                                                                    //
});                                                                    //
                                                                       //
if (Meteor.isClient) {                                                 // 562
  (function () {                                                       //
    Tracker.autorun(function () {                                      // 563
      // We need to keep track of certain data about each grain we can view.
      // TODO(cleanup): Do these in GrainView to avoid spurious resubscribes.
      var grains = globalGrains.get();                                 // 566
      grains.forEach(function (grain) {                                // 567
        grain.depend();                                                // 568
        var grainId = grain.grainId();                                 // 569
        if (grainId) {                                                 // 570
          Meteor.subscribe("grainTopBar", grainId);                    // 571
          if (grain.isOwner()) {                                       // 572
            Meteor.subscribe("packageByGrainId", grainId);             // 573
          }                                                            //
                                                                       //
          var token = grain.token();                                   // 576
          if (token) {                                                 // 577
            Meteor.subscribe("tokenInfo", token);                      // 578
          }                                                            //
        }                                                              //
      });                                                              //
    });                                                                //
                                                                       //
    Template.layout.events({                                           // 584
      "click .incognito-button": function (event) {                    // 585
        console.log(event);                                            // 586
        var grains = globalGrains.get();                               // 587
        var token = event.currentTarget.getAttribute("data-token");    // 588
        if (token) {                                                   // 589
          grains.forEach(function (grain) {                            // 590
            if (grain.token() == token) {                              // 591
              grain.doNotRevealIdentity();                             // 592
            }                                                          //
          });                                                          //
        } else {                                                       //
          console.error("Interstitial prompt answered, but no token present?");
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    var promptNewTitle = function () {                                 // 601
      var grain = getActiveGrain(globalGrains.get());                  // 602
      if (grain) {                                                     // 603
        var _prompt = "Set new title:";                                // 604
        if (!grain.isOwner()) {                                        // 605
          _prompt = "Set a new personal title: (does not change the owner's title for this grain)";
        }                                                              //
                                                                       //
        var title = window.prompt(_prompt, grain.title());             // 609
        if (title) {                                                   // 610
          grain.setTitle(title);                                       // 611
        }                                                              //
      }                                                                //
    };                                                                 //
                                                                       //
    Template.grainTitle.events({                                       // 616
      click: function (event) {                                        // 617
        promptNewTitle();                                              // 618
      },                                                               //
                                                                       //
      keydown: function (event) {                                      // 621
        if (event.keyCode === 13 || event.keyCode === 32) {            // 622
          // Allow space or enter to trigger renaming the grain - Firefox doesn't treat enter on the
          // focused element as click().                               //
          promptNewTitle();                                            // 625
          event.preventDefault();                                      // 626
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainDeleteButton.events({                                // 631
      "click button": function (event) {                               // 632
        var grains = globalGrains.get();                               // 633
        var activeIndex = activeGrainIndex(grains);                    // 634
        var activeGrain = grains[activeIndex];                         // 635
        var newActiveIndex = activeIndex == grains.length - 1 ? activeIndex - 1 : activeIndex;
        if (activeGrain.isOwner()) {                                   // 637
          if (window.confirm("Really delete this grain?")) {           // 638
            Meteor.call("deleteGrain", activeGrain.grainId());         // 639
            // TODO: extract globalGrains into a class that has a "close" method for closing the active view
            activeGrain.destroy();                                     // 641
            if (grains.length == 1) {                                  // 642
              globalGrains.set([]);                                    // 643
              Router.go("root"); // will redirect to the right place   // 644
            } else {                                                   //
                grains.splice(activeIndex, 1);                         // 646
                grains[newActiveIndex].setActive(true);                // 647
                globalGrains.set(grains);                              // 648
                Router.go(grains[newActiveIndex].route());             // 649
              }                                                        //
          }                                                            //
        } else {                                                       //
          if (window.confirm("Really forget this grain?")) {           // 653
            var identityId = activeGrain.identityId();                 // 654
            if (identityId) {                                          // 655
              Meteor.call("forgetGrain", activeGrain.grainId(), identityId);
            }                                                          //
            // TODO: extract globalGrains into a class that has a "close" method for closing the active view
            activeGrain.destroy();                                     // 659
            if (grains.length == 1) {                                  // 660
              globalGrains.set([]);                                    // 661
              Router.go("root"); // will redirect to the right place   // 662
            } else {                                                   //
                grains.splice(activeIndex, 1);                         // 664
                grains[newActiveIndex].setActive(true);                // 665
                globalGrains.set(grains);                              // 666
                Router.go(grains[newActiveIndex].route());             // 667
              }                                                        //
          }                                                            //
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainDebugLogButton.events({                              // 674
      "click button": function (event) {                               // 675
        this.reset();                                                  // 676
        var activeGrain = getActiveGrain(globalGrains.get());          // 677
        window.open("/grainlog/" + activeGrain.grainId(), "_blank", "menubar=no,status=no,toolbar=no,width=700,height=700");
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainBackupButton.onCreated(function () {                 // 683
      this.isLoading = new ReactiveVar(false);                         // 684
    });                                                                //
                                                                       //
    Template.grainBackupButton.helpers({                               // 687
      isLoading: function () {                                         // 688
        return Template.instance().isLoading.get();                    // 689
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainBackupButton.events({                                // 693
      "click button": function (event, template) {                     // 694
        template.isLoading.set(true);                                  // 695
        this.reset();                                                  // 696
        var activeGrain = getActiveGrain(globalGrains.get());          // 697
        Meteor.call("backupGrain", activeGrain.grainId(), function (err, id) {
          template.isLoading.set(false);                               // 699
          if (err) {                                                   // 700
            alert("Backup failed: " + err); // TODO(someday): make this better UI
          } else {                                                     //
              // Firefox for some reason decides to kill all websockets when we try to download the file
              // by navigating there. So we're left doing a dirty hack to get around the popup blocker.
              var isFirefox = typeof InstallTrigger !== "undefined";   // 705
                                                                       //
              if (isFirefox) {                                         // 707
                var save = document.createElement("a");                // 708
                save.href = "/downloadBackup/" + id;                   // 709
                                                                       //
                save.download = activeGrain.title() + ".zip";          // 711
                var _event = document.createEvent("MouseEvents");      // 712
                _event.initMouseEvent("click", true, false, window, 0, 0, 0, 0, 0, false, false, false, false, 0, null);
                save.dispatchEvent(_event);                            // 717
              } else {                                                 //
                window.location = "/downloadBackup/" + id;             // 719
              }                                                        //
            }                                                          //
        });                                                            //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainRestartButton.events({                               // 726
      "click button": function (event) {                               // 727
        this.reset();                                                  // 728
        var activeGrain = getActiveGrain(globalGrains.get());          // 729
        var grainId = activeGrain.grainId();                           // 730
                                                                       //
        Meteor.call("shutdownGrain", grainId, function (err) {         // 732
          if (err) {                                                   // 733
            alert("Restart failed: " + err); // TODO(someday): make this better UI
          } else {                                                     //
              var _frames = document.getElementsByClassName("grain-frame");
              for (var i = 0; i < _frames.length; i++) {               // 737
                var frame = _frames[i];                                // 738
                if (frame.dataset.grainid == grainId) {                // 739
                  frame.src = frame.src;                               // 740
                }                                                      //
              }                                                        //
            }                                                          //
        });                                                            //
      }                                                                //
    });                                                                //
                                                                       //
    function selectElementContents(element) {                          // 748
      if (document.body.createTextRange) {                             // 749
        var range = document.body.createTextRange();                   // 750
        range.moveToElementText(element);                              // 751
        range.select();                                                // 752
      } else if (window.getSelection) {                                //
        var selection = window.getSelection();                         // 754
        var range = document.createRange();                            // 755
        range.selectNodeContents(element);                             // 756
        selection.removeAllRanges();                                   // 757
        selection.addRange(range);                                     // 758
      }                                                                //
    }                                                                  //
                                                                       //
    function selectTargetContents(event) {                             // 762
      event.preventDefault();                                          // 763
      selectElementContents(event.currentTarget);                      // 764
    }                                                                  //
                                                                       //
    Template.grainApiTokenPopup.events({                               // 767
      "click .copy-me": selectTargetContents,                          // 768
      "focus .copy-me": selectTargetContents,                          // 769
      "submit .newApiToken": function (event) {                        // 770
        event.preventDefault();                                        // 771
        var activeGrain = getActiveGrain(globalGrains.get());          // 772
        var grainId = activeGrain.grainId();                           // 773
        activeGrain.setGeneratedApiToken("pending");                   // 774
        var roleList = document.getElementById("api-token-role");      // 775
        // TODO(cleanup): avoid using global ids; select a child of the current template instead
        var assignment = { allAccess: null };                          // 777
        if (roleList && roleList.selectedIndex > 0) {                  // 778
          assignment = { roleId: roleList.selectedIndex - 1 };         // 779
        }                                                              //
                                                                       //
        Meteor.call("newApiToken", { identityId: activeGrain.identityId() }, grainId, document.getElementById("api-token-petname").value, assignment, { webkey: { forSharing: false } }, function (error, result) {
          if (error) {                                                 // 786
            activeGrain.setGeneratedApiToken(undefined);               // 787
            window.alert("Failed to create token.\n" + error);         // 788
            console.error(error.stack);                                // 789
          } else {                                                     //
            activeGrain.setGeneratedApiToken(result.endpointUrl + "#" + result.token);
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click #resetApiToken": function (event) {                       // 796
        var activeGrain = getActiveGrain(globalGrains.get());          // 797
        activeGrain.setGeneratedApiToken(undefined);                   // 798
      },                                                               //
                                                                       //
      "click button.revoke-token": function (event) {                  // 801
        Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { revoked: true });
      },                                                               //
                                                                       //
      "click .token-petname": function (event) {                       // 806
        // TODO(soon): Find a less-annoying way to get this input, perhaps by allowing the user
        //   to edit the petname in place.                             //
        var petname = window.prompt("Set new label:", this.petname);   // 809
        if (petname) {                                                 // 810
          Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { petname: petname });
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainSharePopup.events({                                  // 817
      "click .copy-me": selectTargetContents,                          // 818
      "focus .copy-me": selectTargetContents,                          // 819
      "click #share-grain-popup-closer": function (event) {            // 820
        Session.set("show-share-grain", false);                        // 821
      },                                                               //
                                                                       //
      "click button.who-has-access": function (event, instance) {      // 824
        event.preventDefault();                                        // 825
        var closer = globalTopbar.addItem({                            // 826
          name: "who-has-access",                                      // 827
          template: Template.whoHasAccess,                             // 828
          popupTemplate: Template.whoHasAccessPopup,                   // 829
          data: new ReactiveVar(instance.data),                        // 830
          startOpen: true,                                             // 831
          onDismiss: function () {                                     // 832
            return "remove";                                           // 833
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click #privatize-grain": function (event) {                     // 838
        Meteor.call("privatizeGrain", getActiveGrain(globalGrains.get()).grainId());
      },                                                               //
                                                                       //
      "click a.open-non-anonymously": function (event) {               // 842
        event.preventDefault();                                        // 843
        globalTopbar.closePopup();                                     // 844
        getActiveGrain(globalGrains.get()).reset();                    // 845
        getActiveGrain(globalGrains.get()).openSession();              // 846
      }                                                                //
    });                                                                //
                                                                       //
    Template.shareWithOthers.onRendered(function () {                  // 850
      this.find("[role=tab]").focus();                                 // 851
    });                                                                //
                                                                       //
    var activateTargetTab = function (event, instance) {               // 854
      // Deactivate all tabs and all tab panels.                       //
      instance.findAll("ul[role=tablist]>li[role=tab]").forEach(function (element) {
        element.setAttribute("aria-selected", false);                  // 857
      });                                                              //
                                                                       //
      instance.findAll(".tabpanel").forEach(function (element) {       // 860
        element.setAttribute("aria-hidden", true);                     // 861
      });                                                              //
                                                                       //
      // Activate the tab header the user selected.                    //
      event.currentTarget.setAttribute("aria-selected", true);         // 865
      // Show the corresponding tab panel.                             //
      var idToShow = event.currentTarget.getAttribute("aria-controls");
      var tabPanelToShow = instance.find("#" + idToShow);              // 868
      tabPanelToShow.setAttribute("aria-hidden", false);               // 869
    };                                                                 //
                                                                       //
    Template.shareWithOthers.events({                                  // 872
      "click #send-invite-tab-header": activateTargetTab,              // 873
      "click #shareable-link-tab-header": activateTargetTab,           // 874
      "keydown [role=tab]": function (event, template) {               // 875
        if (event.keyCode == 38 || event.keyCode == 40) {              // 876
          // up and down arrows                                        //
          event.preventDefault();                                      // 877
        }                                                              //
                                                                       //
        var $focus = $(template.find(":focus"));                       // 880
        var $items = template.$("[role=tab]:visible");                 // 881
        var focusIndex = $items.index($focus);                         // 882
        var newFocusIndex = undefined;                                 // 883
        if (event.keyCode == 37) {                                     // 884
          // left arrow                                                //
          event.preventDefault();                                      // 885
          newFocusIndex = focusIndex - 1;                              // 886
          if (newFocusIndex == -1) {                                   // 887
            newFocusIndex = $items.length - 1;                         // 888
          }                                                            //
        } else if (event.keyCode == 39) {                              //
          // right arrow                                               //
          event.preventDefault();                                      // 891
          newFocusIndex = focusIndex + 1;                              // 892
          if (newFocusIndex >= $items.length) {                        // 893
            newFocusIndex = 0;                                         // 894
          }                                                            //
        } else if (event.keyCode == 13) {                              //
          // Enter key                                                 //
          event.preventDefault();                                      // 897
          activateTargetTab(event, template);                          // 898
        }                                                              //
                                                                       //
        if (newFocusIndex != null) {                                   // 901
          $items.attr("tabindex", "-1");                               // 902
          var $newFocus = $($items[newFocusIndex]);                    // 903
          $newFocus.attr("tabindex", "0");                             // 904
          $newFocus.focus();                                           // 905
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.shareableLinkTab.events({                                 // 910
      "change .share-token-role": function (event, instance) {         // 911
        var success = instance.completionState.get().success;          // 912
        if (success) {                                                 // 913
          var roleList = event.target;                                 // 914
          var assignment = undefined;                                  // 915
          if (roleList) {                                              // 916
            assignment = { roleId: roleList.selectedIndex };           // 917
          } else {                                                     //
            assignment = { none: null };                               // 919
          }                                                            //
                                                                       //
          Meteor.call("updateApiToken", success.id, { roleAssignment: assignment }, function (error) {
            if (error) {                                               // 923
              console.error(error.stack);                              // 924
            }                                                          //
          });                                                          //
        }                                                              //
      },                                                               //
                                                                       //
      "change .label": function (event, instance) {                    // 930
        var success = instance.completionState.get().success;          // 931
        if (success) {                                                 // 932
          var label = event.target.value;                              // 933
          Meteor.call("updateApiToken", success.id, { petname: label }, function (error) {
            if (error) {                                               // 935
              console.error(error.stack);                              // 936
            }                                                          //
          });                                                          //
        }                                                              //
      },                                                               //
                                                                       //
      "submit form.new-share-token": function (event, instance) {      // 942
        event.preventDefault();                                        // 943
        if (!instance.completionState.get().clear) {                   // 944
          return;                                                      // 945
        }                                                              //
                                                                       //
        var currentGrain = getActiveGrain(globalGrains.get());         // 948
        var grainId = currentGrain.grainId();                          // 949
        var roleList = event.target.getElementsByClassName("share-token-role")[0];
        var assignment = undefined;                                    // 951
        if (roleList) {                                                // 952
          assignment = { roleId: roleList.selectedIndex };             // 953
        } else {                                                       //
          assignment = { none: null };                                 // 955
        }                                                              //
                                                                       //
        instance.completionState.set({ pending: true });               // 958
        Meteor.call("newApiToken", { identityId: currentGrain.identityId() }, grainId, event.target.getElementsByClassName("label")[0].value, assignment, { webkey: { forSharing: true } }, function (error, result) {
          if (error) {                                                 // 963
            console.error(error.stack);                                // 964
          } else {                                                     //
            result.url = getOrigin() + "/shared/" + result.token;      // 966
            instance.completionState.set({ success: result });         // 967
            // On the next render, .copy-me will exist, and we should focus it then.
            Meteor.defer(function () {                                 // 969
              var element = instance.find(".copy-me");                 // 970
              element.focus();                                         // 971
              selectElementContents(element);                          // 972
            });                                                        //
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click .reset-share-token": function (event, instance) {         // 978
        instance.completionState.set({ clear: true });                 // 979
        instance.find("form").reset();                                 // 980
        instance.find("form option[data-default-selected=true]").selected = true;
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainPowerboxOfferPopup.events({                          // 985
      "click button.dismiss": function (event) {                       // 986
        var data = Template.instance().data.get();                     // 987
        data.onDismiss();                                              // 988
      },                                                               //
                                                                       //
      "click .copy-me": selectTargetContents,                          // 991
      "focus .copy-me": selectTargetContents                           // 992
    });                                                                //
                                                                       //
    Template.grainSharePopup.helpers({                                 // 995
      incognito: function () {                                         // 996
        return !Accounts.getCurrentIdentityId();                       // 997
      },                                                               //
                                                                       //
      currentTokenUrl: function () {                                   // 1000
        return getOrigin() + "/shared/" + getActiveGrain(globalGrains.get()).token();
      },                                                               //
                                                                       //
      currentGrain: function () {                                      // 1004
        return getActiveGrain(globalGrains.get());                     // 1005
      }                                                                //
    });                                                                //
                                                                       //
    Template.requestAccess.onCreated(function () {                     // 1009
      this._status = new ReactiveVar({ showButton: true });            // 1010
      this._grain = getActiveGrain(globalGrains.get());                // 1011
    });                                                                //
                                                                       //
    Template.requestAccess.events({                                    // 1014
      "click button.request-access": function (event, instance) {      // 1015
        instance._status.set({ chooseIdentity: true });                // 1016
      }                                                                //
    });                                                                //
                                                                       //
    Template.requestAccess.helpers({                                   // 1020
      status: function () {                                            // 1021
        return Template.instance()._status.get();                      // 1022
      },                                                               //
                                                                       //
      chooseIdentityText: function () {                                // 1025
        if (SandstormDb.getUserIdentityIds(Meteor.user()).length > 1) {
          return "Please select an identity with which to request access.";
        } else {                                                       //
          return "To confirm, please click on your identity below.";   // 1029
        }                                                              //
      },                                                               //
                                                                       //
      identityPickerData: function () {                                // 1033
        var identities = SandstormDb.getUserIdentityIds(Meteor.user()).map(function (id) {
          return globalDb.getIdentity(id);                             //
        });                                                            //
        var instance = Template.instance();                            // 1036
        function onPicked(identityId) {                                // 1037
          var grainId = instance._grain.grainId();                     // 1038
          Meteor.call("requestAccess", getOrigin(), grainId, identityId, function (error, result) {
            if (error) {                                               // 1040
              instance._status.set({ error: error });                  // 1041
            } else {                                                   //
              instance._status.set({ success: true });                 // 1043
              instance.subscribe("requestingAccess", grainId, identityId);
              instance.autorun(function () {                           // 1045
                var granted = GrantedAccessRequests.findOne({ grainId: grainId,
                  identityId: identityId });                           // 1047
                if (granted) {                                         // 1048
                  instance._grain.reset(granted.identityId);           // 1049
                  instance._grain.openSession();                       // 1050
                }                                                      //
              });                                                      //
            }                                                          //
          });                                                          //
                                                                       //
          instance._status.set({ waiting: true });                     // 1056
        }                                                              //
                                                                       //
        return {                                                       // 1059
          identities: identities,                                      // 1060
          onPicked: onPicked                                           // 1061
        };                                                             //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainView.helpers({                                       // 1066
      unpackedGrainState: function () {                                // 1067
        return mapGrainStateToTemplateData(this);                      // 1068
      },                                                               //
                                                                       //
      identityPickerData: function () {                                // 1071
        var identities = SandstormDb.getUserIdentityIds(Meteor.user()).map(function (id) {
          var identity = Meteor.users.findOne({ _id: id });            // 1073
          if (identity) {                                              // 1074
            SandstormDb.fillInProfileDefaults(identity);               // 1075
            SandstormDb.fillInIntrinsicName(identity);                 // 1076
            SandstormDb.fillInPictureUrl(identity);                    // 1077
            return identity;                                           // 1078
          }                                                            //
        });                                                            //
                                                                       //
        var grain = getActiveGrain(globalGrains.get());                // 1082
        return { identities: identities,                               // 1083
          onPicked: function (identityId) {                            // 1084
            grain.revealIdentity(identityId);                          // 1084
          } };                                                         //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grain.helpers({                                           // 1088
      currentGrain: function () {                                      // 1089
        return getActiveGrain(globalGrains.get());                     // 1090
      },                                                               //
                                                                       //
      isOwner: function () {                                           // 1093
        var current = getActiveGrain(globalGrains.get());              // 1094
        return current && current.isOwner();                           // 1095
      },                                                               //
                                                                       //
      grainSize: function () {                                         // 1098
        var current = getActiveGrain(globalGrains.get());              // 1099
        var sizeBytes = current && current.size();                     // 1100
        if (sizeBytes !== undefined) {                                 // 1101
          return "(" + prettySize(sizeBytes) + ")";                    // 1102
        } else {                                                       //
          return "";                                                   // 1104
        }                                                              //
      },                                                               //
                                                                       //
      setGrainWindowTitle: function () {                               // 1108
        var current = getActiveGrain(globalGrains.get());              // 1109
        if (current) {                                                 // 1110
          current.updateDocumentTitle();                               // 1111
        }                                                              //
      },                                                               //
                                                                       //
      displayWebkeyButton: function () {                               // 1115
        // TODO: figure out what this should do                        //
        return Meteor.userId() || !this.oldSharingModel;               // 1117
      },                                                               //
                                                                       //
      showPowerboxOffer: function () {                                 // 1120
        var current = getActiveGrain(globalGrains.get());              // 1121
        return current && current.showPowerboxOffer();                 // 1122
      },                                                               //
                                                                       //
      powerboxOfferData: function () {                                 // 1125
        var current = getActiveGrain(globalGrains.get());              // 1126
        return current && current.powerboxOfferData();                 // 1127
      },                                                               //
                                                                       //
      showPowerboxRequest: function () {                               // 1130
        var current = getActiveGrain(globalGrains.get());              // 1131
        return current && current.showPowerboxRequest();               // 1132
      },                                                               //
                                                                       //
      powerboxRequestData: function () {                               // 1135
        var current = getActiveGrain(globalGrains.get());              // 1136
        return current && current.powerboxRequestData();               // 1137
      },                                                               //
                                                                       //
      cancelPowerboxRequest: function () {                             // 1140
        return function () {                                           // 1141
          var current = getActiveGrain(globalGrains.get());            // 1142
          current.setPowerboxRequest(undefined);                       // 1143
          return "remove";                                             // 1144
        };                                                             //
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainTitle.helpers({                                      // 1149
      title: function () {                                             // 1150
        var grain = getActiveGrain(globalGrains.get());                // 1151
        return grain && grain.title() || "Untitled grain";             // 1152
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainApiTokenPopup.helpers({                              // 1156
      displayToken: function () {                                      // 1157
        return !this.revoked && !this.expiresIfUnused && !this.parentToken;
      },                                                               //
                                                                       //
      existingTokens: function () {                                    // 1161
        var current = getActiveGrain(globalGrains.get());              // 1162
        return current && ApiTokens.find({                             // 1163
          grainId: current.grainId(),                                  // 1164
          identityId: current.identityId(),                            // 1165
          forSharing: { $ne: true },                                   // 1166
          $or: [{ owner: { webkey: null } }, { owner: { $exists: false } }],
          expiresIfUnused: null                                        // 1171
        });                                                            //
      },                                                               //
                                                                       //
      generatedApiToken: function () {                                 // 1175
        var current = getActiveGrain(globalGrains.get());              // 1176
        return current && current.generatedApiToken();                 // 1177
      },                                                               //
                                                                       //
      generatedApiTokenPending: function () {                          // 1180
        var current = getActiveGrain(globalGrains.get());              // 1181
        return (current && current.generatedApiToken()) == "pending";  // 1182
      },                                                               //
                                                                       //
      viewInfo: function () {                                          // 1185
        var current = getActiveGrain(globalGrains.get());              // 1186
        return current && current.viewInfo();                          // 1187
      }                                                                //
    });                                                                //
                                                                       //
    Template.whoHasAccessPopup.onCreated(function () {                 // 1191
      var _this = this;                                                // 1192
      this.subscribe("contactProfiles");                               // 1193
      var currentGrain = getActiveGrain(globalGrains.get());           // 1194
      _this.identityId = currentGrain.identityId();                    // 1195
      _this.grainId = currentGrain.grainId();                          // 1196
      _this.transitiveShares = new ReactiveVar(null);                  // 1197
      _this.downstreamTokensById = new ReactiveVar({});                // 1198
      this.resetTransitiveShares = function () {                       // 1199
        Meteor.call("transitiveShares", _this.identityId, _this.grainId, function (error, downstream) {
          if (error) {                                                 // 1202
            console.error(error.stack);                                // 1203
          } else {                                                     //
            (function () {                                             //
              var downstreamTokensById = {};                           // 1205
              var sharesByRecipient = {};                              // 1206
              downstream.forEach(function (token) {                    // 1207
                downstreamTokensById[token._id] = token;               // 1208
                if (Match.test(token.owner, { user: Match.ObjectIncluding({ identityId: String }) })) {
                  var recipient = token.owner.user.identityId;         // 1210
                  if (!sharesByRecipient[recipient]) {                 // 1211
                    sharesByRecipient[recipient] = {                   // 1212
                      recipient: recipient,                            // 1213
                      dedupedShares: [],                               // 1214
                      allShares: []                                    // 1215
                    };                                                 //
                  }                                                    //
                                                                       //
                  sharesByRecipient[recipient].allShares.push(token);  // 1219
                  var dedupedShares = sharesByRecipient[recipient].dedupedShares;
                  if (!dedupedShares.some(function (share) {           // 1221
                    return share.identityId === token.identityId;      //
                  })) {                                                //
                    dedupedShares.push(token);                         // 1222
                  }                                                    //
                }                                                      //
              });                                                      //
                                                                       //
              var result = _.values(sharesByRecipient);                // 1227
              if (result.length == 0) {                                // 1228
                result = { empty: true };                              // 1229
              }                                                        //
                                                                       //
              _this.transitiveShares.set(result);                      // 1232
              _this.downstreamTokensById.set(downstreamTokensById);    // 1233
            })();                                                      //
          }                                                            //
        });                                                            //
      };                                                               //
                                                                       //
      this.resetTransitiveShares();                                    // 1238
    });                                                                //
                                                                       //
    Template.whoHasAccessPopup.events({                                // 1241
      "change .share-token-role": function (event, instance) {         // 1242
        var roleList = event.target;                                   // 1243
        var assignment = undefined;                                    // 1244
        if (roleList) {                                                // 1245
          assignment = { roleId: roleList.selectedIndex };             // 1246
        } else {                                                       //
          assignment = { none: null };                                 // 1248
        }                                                              //
                                                                       //
        Meteor.call("updateApiToken", roleList.getAttribute("data-token-id"), { roleAssignment: assignment }, function (error) {
          if (error) {                                                 // 1253
            console.error(error.stack);                                // 1254
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click button.revoke-token": function (event, instance) {        // 1259
        Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { revoked: true });
        instance.resetTransitiveShares();                              // 1262
      },                                                               //
                                                                       //
      "click button.revoke-access": function (event, instance) {       // 1265
        var recipient = event.currentTarget.getAttribute("data-recipient");
        var transitiveShares = instance.transitiveShares.get();        // 1267
        var tokensById = instance.downstreamTokensById.get();          // 1268
        var recipientShares = _.findWhere(transitiveShares, { recipient: recipient });
        var currentIdentityId = Accounts.getCurrentIdentityId();       // 1270
        var recipientTokens = _.where(recipientShares.allShares, { identityId: currentIdentityId });
                                                                       //
        // Two cases:                                                  //
        // 1. All of the tokens are direct shares. Easy. Just revoke them.
        // 2. Some of the links are child tokens. For each, we walk up the chain of parents to the
        //    root. Then we walk downwards to see whether any other identities would be immediately
        //    affected by revoking those roots. We then collect that data and display it in a
        //    confirmation dialog.                                     //
                                                                       //
        var directTokens = [];                                         // 1280
        // Direct shares from the current identity to the recipient.   //
                                                                       //
        var rootTokens = [];                                           // 1283
        // Roots of chains of child tokens starting at the current identity and leading to the
        // recipient.                                                  //
                                                                       //
        recipientTokens.forEach(function (token) {                     // 1287
          if (token.parentToken) {                                     // 1288
            var currentToken = token;                                  // 1289
            do {                                                       // 1290
              currentToken = tokensById[token.parentToken];            // 1291
            } while (currentToken.parentToken);                        //
                                                                       //
            rootTokens.push(currentToken);                             // 1294
          } else {                                                     //
            directTokens.push(token);                                  // 1296
          }                                                            //
        });                                                            //
                                                                       //
        if (rootTokens.length > 0) {                                   // 1300
          var _ret5 = (function () {                                   //
            // Some of the shares are not direct.                      //
                                                                       //
            var tokensByParent = {};                                   // 1303
            for (var id in babelHelpers.sanitizeForInObject(tokensById)) {
              var token = tokensById[id];                              // 1305
              if (token.parentToken) {                                 // 1306
                if (!tokensByParent[token.parentToken]) {              // 1307
                  tokensByParent[token.parentToken] = [];              // 1308
                }                                                      //
                                                                       //
                tokensByParent[token.parentToken].push(token);         // 1311
              }                                                        //
            }                                                          //
                                                                       //
            var otherAffectedIdentities = {};                          // 1315
            var tokenStack = rootTokens.slice(0);                      // 1316
            while (tokenStack.length > 0) {                            // 1317
              var current = tokenStack.pop();                          // 1318
              if (Match.test(current.owner, { user: Match.ObjectIncluding({ identityId: String }) })) {
                if (current.owner.user.identityId != recipient) {      // 1321
                  otherAffectedIdentities[current.owner.user.identityId] = true;
                }                                                      //
              } else {                                                 //
                if (tokensByParent[current._id]) {                     // 1325
                  var children = tokensByParent[current._id];          // 1326
                  children.forEach(function (child) {                  // 1327
                    tokenStack.push(child);                            // 1328
                  });                                                  //
                }                                                      //
              }                                                        //
            }                                                          //
                                                                       //
            var recipientIdentity = ContactProfiles.findOne({ _id: recipient });
            var recipientName = recipientIdentity && recipientIdentity.profile.name || "Unknown User";
            var singular = rootTokens.length == 1;                     // 1337
                                                                       //
            var tokenLabels = _.pluck(rootTokens, "petname").map(function (petname) {
              return petname || "Unlabeled Link";                      //
            }).map(function (petname) {                                //
              return "\"" + petname + "\"";                            //
            }).join(", ");                                             //
                                                                       //
            var confirmText = "This will revoke the following sharing link" + (singular ? "" : "s") + ":\n\n    " + tokenLabels + "\n\n";
                                                                       //
            var othersNote = "(No signed-in user other than " + recipientName + " has opened " + (singular ? "this link" : "these links") + " yet.)";
            if (Object.keys(otherAffectedIdentities).length > 0) {     // 1349
              var othersNames = Object.keys(otherAffectedIdentities).map(function (identityId) {
                return ContactProfiles.findOne({ _id: identityId });   //
              }).map(function (identity) {                             //
                return identity && identity.profile.name || "Unknown User";
              }).join(", ");                                           //
              othersNote = (singular ? "This link has" : "These links have") + " also been opened by:\n\n    " + othersNames;
            }                                                          //
                                                                       //
            if (window.confirm(confirmText + othersNote)) {            // 1358
              rootTokens.forEach(function (token) {                    // 1359
                Meteor.call("updateApiToken", token._id, { revoked: true });
              });                                                      //
            } else {                                                   //
              // Cancel.                                               //
              return {                                                 // 1364
                v: undefined                                           //
              };                                                       //
            }                                                          //
          })();                                                        //
                                                                       //
          if (typeof _ret5 === "object") return _ret5.v;               //
        }                                                              //
                                                                       //
        directTokens.forEach(function (token) {                        // 1368
          Meteor.call("updateApiToken", token._id, { revoked: true });
        });                                                            //
                                                                       //
        instance.resetTransitiveShares();                              // 1372
      },                                                               //
                                                                       //
      "click .token-petname": function (event, instance) {             // 1375
        // TODO(soon): Find a less-annoying way to get this input, perhaps by allowing the user
        //   to edit the petname in place.                             //
        var petname = window.prompt("Set new label:", this.petname);   // 1378
        if (petname) {                                                 // 1379
          Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { petname: petname });
        }                                                              //
                                                                       //
        instance.resetTransitiveShares();                              // 1384
      }                                                                //
    });                                                                //
                                                                       //
    function isEmptyPermissionSet(permissionSet) {                     // 1388
      if (!permissionSet) {                                            // 1389
        return true;                                                   // 1390
      }                                                                //
                                                                       //
      for (var ii = 0; ii < permissionSet.length; ++ii) {              // 1393
        if (permissionSet[ii]) {                                       // 1394
          return false;                                                // 1395
        }                                                              //
      }                                                                //
                                                                       //
      return true;                                                     // 1399
    }                                                                  //
                                                                       //
    Template.whoHasAccessPopup.helpers({                               // 1402
      existingShareTokens: function () {                               // 1403
        if (Meteor.userId()) {                                         // 1404
          return ApiTokens.find({                                      // 1405
            grainId: Template.instance().grainId,                      // 1406
            identityId: Template.instance().identityId,                // 1407
            forSharing: true,                                          // 1408
            $or: [{ owner: { webkey: null } }, { owner: { $exists: false } }]
          }).fetch();                                                  //
        }                                                              //
      },                                                               //
                                                                       //
      isCurrentIdentity: function () {                                 // 1417
        if (this.identityId === Accounts.getCurrentIdentityId()) {     // 1418
          return true;                                                 // 1419
        }                                                              //
      },                                                               //
                                                                       //
      getPetname: function () {                                        // 1423
        if (this.petname) {                                            // 1424
          return this.petname;                                         // 1425
        } else {                                                       //
          return "Unlabeled Link";                                     // 1427
        }                                                              //
      },                                                               //
                                                                       //
      displayName: function (identityId) {                             // 1431
        var identity = ContactProfiles.findOne({ _id: identityId });   // 1432
        if (!identity) {                                               // 1433
          identity = Meteor.users.findOne({ _id: identityId });        // 1434
        }                                                              //
                                                                       //
        if (identity) {                                                // 1437
          return identity.profile.name;                                // 1438
        } else {                                                       //
          return "Unknown User (" + identityId.slice(0, 16) + ")";     // 1440
        }                                                              //
      },                                                               //
                                                                       //
      transitiveShares: function () {                                  // 1444
        return Template.instance().transitiveShares.get();             // 1445
      },                                                               //
                                                                       //
      indexedRoles: function () {                                      // 1448
        var result = [];                                               // 1449
        var instance = Template.instance();                            // 1450
        var currentGrain = getActiveGrain(globalGrains.get());         // 1451
        var roles = currentGrain.viewInfo().roles;                     // 1452
        for (var ii = 0; ii < roles.length; ++ii) {                    // 1453
          result.push({ idx: ii, title: roles[ii].title, verbPhrase: roles[ii].verbPhrase });
        }                                                              //
                                                                       //
        return result;                                                 // 1457
      },                                                               //
                                                                       //
      roleText: function () {                                          // 1460
        if (this.verbPhrase) {                                         // 1461
          return this.verbPhrase.defaultText;                          // 1462
        } else {                                                       //
          return "is " + this.title.defaultText;                       // 1464
        }                                                              //
      },                                                               //
                                                                       //
      hasCustomRole: function (token) {                                // 1468
        var role = token.roleAssignment;                               // 1469
        if ("roleId" in role && isEmptyPermissionSet(role.addPermissions) && isEmptyPermissionSet(role.removePermissions)) {
          return false;                                                // 1473
        }                                                              //
                                                                       //
        return true;                                                   // 1476
      },                                                               //
                                                                       //
      hasCurrentRole: function (token) {                               // 1479
        var role = token.roleAssignment;                               // 1480
        if ("roleId" in role && role.roleId == this.idx && isEmptyPermissionSet(role.addPermissions) && isEmptyPermissionSet(role.removePermissions)) {
          return true;                                                 // 1484
        }                                                              //
                                                                       //
        return false;                                                  // 1487
      },                                                               //
                                                                       //
      displayToken: function () {                                      // 1490
        return !this.revoked && !this.expiresIfUnused && !this.parentToken;
      },                                                               //
                                                                       //
      viewInfo: function () {                                          // 1494
        var activeGrain = getActiveGrain(globalGrains.get());          // 1495
        return activeGrain && activeGrain.viewInfo();                  // 1496
      }                                                                //
    });                                                                //
                                                                       //
    Template.selectRole.helpers({                                      // 1500
      roleText: function () {                                          // 1501
        if (this.verbPhrase) {                                         // 1502
          return this.verbPhrase.defaultText;                          // 1503
        } else {                                                       //
          return "is " + this.title.defaultText;                       // 1505
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    Template.shareableLinkTab.onCreated(function () {                  // 1510
      this.completionState = new ReactiveVar({ clear: true });         // 1511
    });                                                                //
                                                                       //
    Template.emailInviteTab.onCreated(function () {                    // 1514
      this.completionState = new ReactiveVar({ clear: true });         // 1515
      this.contacts = new ReactiveVar([]);                             // 1516
      this.grain = getActiveGrain(globalGrains.get());                 // 1517
    });                                                                //
                                                                       //
    Template.emailInviteTab.onDestroyed(function () {                  // 1520
      Session.set("share-grain-" + Template.instance().grain.grainId(), undefined);
    });                                                                //
                                                                       //
    Template.shareableLinkTab.helpers({                                // 1524
      completionState: function () {                                   // 1525
        var instance = Template.instance();                            // 1526
        return instance.completionState.get();                         // 1527
      }                                                                //
    });                                                                //
    Template.emailInviteTab.helpers({                                  // 1530
      completionState: function () {                                   // 1531
        var instance = Template.instance();                            // 1532
        return instance.completionState.get();                         // 1533
      },                                                               //
                                                                       //
      contacts: function () {                                          // 1536
        return Template.instance().contacts;                           // 1537
      },                                                               //
                                                                       //
      preselectedIdentityId: function () {                             // 1540
        return Session.get("share-grain-" + Template.instance().grain.grainId());
      }                                                                //
    });                                                                //
                                                                       //
    Template.emailInviteTab.events({                                   // 1545
      "submit form.email-invite": function (event, instance) {         // 1546
        event.preventDefault();                                        // 1547
        return false;                                                  // 1548
      },                                                               //
                                                                       //
      "click form.email-invite button": function (event, instance) {   // 1551
        event.preventDefault();                                        // 1552
        if (!instance.completionState.get().clear) {                   // 1553
          return;                                                      // 1554
        }                                                              //
                                                                       //
        var grainId = instance.data.grainId;                           // 1557
        var title = instance.data.title;                               // 1558
                                                                       //
        var roleList = instance.find(".share-token-role");             // 1560
        var assignment = undefined;                                    // 1561
        if (roleList) {                                                // 1562
          assignment = { roleId: roleList.selectedIndex };             // 1563
        } else {                                                       //
          assignment = { none: null };                                 // 1565
        }                                                              //
                                                                       //
        var message = instance.find(".personal-message").value;        // 1568
        instance.completionState.set({ pending: true });               // 1569
                                                                       //
        var currentGrain = getActiveGrain(globalGrains.get());         // 1571
                                                                       //
        // HTML-escape the message.                                    //
        var div = document.createElement("div");                       // 1574
        div.appendChild(document.createTextNode(message));             // 1575
        var htmlMessage = div.innerHTML.replace(/\n/g, "<br>");        // 1576
                                                                       //
        var contacts = instance.contacts.get();                        // 1578
        var emails = instance.find("input.emails");                    // 1579
        var emailsValue = emails.value;                                // 1580
        if (emailsValue) {                                             // 1581
          if (emailsValue.match(/.+\@.+\..+/)) {                       // 1582
            contacts.push({                                            // 1583
              _id: emailsValue,                                        // 1584
              profile: {                                               // 1585
                service: "email",                                      // 1586
                name: emailsValue,                                     // 1587
                intrinsicName: emailsValue,                            // 1588
                pictureUrl: "/email.svg"                               // 1589
              },                                                       //
              isDefault: true                                          // 1591
            });                                                        //
            instance.contacts.set(contacts);                           // 1593
            emails.value = "";                                         // 1594
          } else {                                                     //
            instance.completionState.set({ error: "Invalid text in contact input box: " + emailsValue });
            return;                                                    // 1598
          }                                                            //
        }                                                              //
                                                                       //
        Meteor.call("inviteUsersToGrain", getOrigin(), currentGrain.identityId(), grainId, title, assignment, contacts, { text: message, html: htmlMessage }, function (error, result) {
          if (error) {                                                 // 1605
            instance.completionState.set({ error: error.toString() });
          } else {                                                     //
            if (result.failures.length > 0) {                          // 1608
              var _message = "Failed to send to: ";                    // 1609
              for (var ii = 0; ii < result.failures.length; ++ii) {    // 1610
                console.error(result.failures[ii].error);              // 1611
                if (ii != 0) {                                         // 1612
                  _message += ", ";                                    // 1613
                }                                                      //
                                                                       //
                _message += result.failures[ii].contact.profile.name;  // 1616
                var warning = result.failures[ii].warning;             // 1617
                if (warning) {                                         // 1618
                  _message += ". " + warning;                          // 1619
                }                                                      //
              }                                                        //
                                                                       //
              instance.completionState.set({ error: _message });       // 1623
            } else {                                                   //
              instance.completionState.set({ success: "success" });    // 1625
            }                                                          //
          }                                                            //
        });                                                            //
      },                                                               //
                                                                       //
      "click .reset-invite": function (event, instance) {              // 1631
        instance.contacts.set([]);                                     // 1632
        instance.completionState.set({ clear: true });                 // 1633
        instance.find("form").reset();                                 // 1634
        instance.find("form option[data-default-selected=true]").selected = true;
      },                                                               //
                                                                       //
      "click .start-over-invite": function (event, instance) {         // 1638
        instance.completionState.set({ clear: true });                 // 1639
      }                                                                //
    });                                                                //
                                                                       //
    Template.grainPowerboxOfferPopup.helpers({                         // 1643
      powerboxOfferUrl: function () {                                  // 1644
        var offer = Template.instance().data.get().offer;              // 1645
        return offer && offer.url;                                     // 1646
      }                                                                //
    });                                                                //
                                                                       //
    // Send a keep-alive for each grain every now and then.            //
    Meteor.setInterval(function () {                                   // 1651
      var grains = globalGrains.get();                                 // 1652
      if (!grains) return;                                             // 1653
                                                                       //
      // Meteor has an exponential backoff of up to 5 minutes for reconnect. This is unnacceptable
      // for us, since we rely on Sessions being re-established in under 60s.
      if (Meteor.status().status === "waiting") {                      // 1657
        console.log("Sandstorm is trying to reconnect...");            // 1658
        Meteor.reconnect();                                            // 1659
      }                                                                //
                                                                       //
      grains.forEach(function (grain) {                                // 1662
        if (grain.sessionId()) {                                       // 1663
          // TODO(soon):  Investigate what happens in background tabs.  Maybe arrange to re-open the
          //   app if it dies while in the background.                 //
          Meteor.call("keepSessionAlive", grain.sessionId(), function (error, result) {
            // Sessions will automatically resume if possible, otherwise they will refresh.
          });                                                          //
        }                                                              //
      });                                                              //
    }, 60000);                                                         //
                                                                       //
    var memoizedNewApiToken = {};                                      // 1673
    // Maps sha256(JSON.stringify(parameters)) -> {timestamp, promise}
    //                                                                 //
    // This memoizes calls to the Meteor method "newApiToken", so that multiple calls in rapid
    // succession will not create multiple tokens. `parameters` above is an array containing the
    // calls parameters in order (excluding the method name and callback), and `promise` is a promise
    // for the result of the call.                                     //
                                                                       //
    // Message handler for Sandstorm's client-side postMessage API.    //
    Meteor.startup(function () {                                       // 1682
      var messageListener = function (event) {                         // 1683
        if (event.origin === getOrigin()) {                            // 1684
          // Meteor likes to postMessage() to itself sometimes, so we ignore these messages.
          return;                                                      // 1686
        }                                                              //
                                                                       //
        // Look up the grain index of which grain this postmessage came from, so we can map behavior
        // into that particular grain's state                          //
        var grains = globalGrains.get();                               // 1691
        var senderGrainIndex = grainOriginToIndex(grains, event.origin);
        if (senderGrainIndex == -1) {                                  // 1693
          // We got a postMessage from an origin that is not a grain we currently believe is open.
          // Ignore it. (It may be aimed at some other message listener registered elsewhere...)
          return;                                                      // 1696
        }                                                              //
                                                                       //
        var senderGrain = grains[senderGrainIndex];                    // 1699
                                                                       //
        if (event.data.setPath || event.data.setPath === "") {         // 1701
          var path = event.data.setPath || "/";                        // 1702
          check(path, String);                                         // 1703
          check(path.charAt(0), "/");                                  // 1704
          // TODO(security): More sanitization of this path. E.g. reject "/../../".
          senderGrain.setPath(path);                                   // 1706
        } else if (event.data.startSharing) {                          //
          // Allow the current grain to request that the "Share Access" menu be shown.
          // Only show this popup if no other popup is currently active.
          // TODO(security): defend against malicious apps spamming this call, blocking all other UI.
          var currentGrain = getActiveGrain(globalGrains.get());       // 1711
          if (senderGrain === currentGrain && !globalTopbar.isPopupOpen()) {
            globalTopbar.openPopup("share");                           // 1713
          }                                                            //
        } else if (event.data.setTitle || event.data.setTitle === "") {
          senderGrain.setFrameTitle(event.data.setTitle);              // 1716
        } else if (event.data.renderTemplate) {                        //
          var _ret6 = (function () {                                   //
            // Request creation of a single-use template with a privileged API token.
            // Why?  Apps should not be able to obtain capabilities-as-keys to
            // themselves directly, because those can be leaked through an
            // arbitrary bit stream or covert channel.  However, apps often need a
            // way to provide instructions to users to copy/paste with some
            // privileged token contained within.  By providing this templating in
            // the platform, we can ensure that the token is only visible to the
            // shell's origin.                                         //
            var call = event.data.renderTemplate;                      // 1726
            check(call, Object);                                       // 1727
            var rpcId = call.rpcId;                                    // 1728
            try {                                                      // 1729
              check(call, { rpcId: String, template: String, petname: Match.Optional(String),
                roleAssignment: Match.Optional(roleAssignmentPattern),
                forSharing: Match.Optional(Boolean) });                // 1732
            } catch (error) {                                          //
              event.source.postMessage({ rpcId: rpcId, error: error.toString() }, event.origin);
              return {                                                 // 1735
                v: undefined                                           //
              };                                                       //
            }                                                          //
                                                                       //
            var template = call.template;                              // 1738
            var petname = "connected external app";                    // 1739
            if (call.petname) {                                        // 1740
              petname = call.petname;                                  // 1741
            }                                                          //
                                                                       //
            var assignment = { allAccess: null };                      // 1744
            if (call.roleAssignment) {                                 // 1745
              assignment = call.roleAssignment;                        // 1746
            }                                                          //
                                                                       //
            var forSharing = call.forSharing ? call.forSharing : false;
            // Tokens expire by default in 5 minutes from generation date
            var selfDestructDuration = 5 * 60 * 1000;                  // 1751
                                                                       //
            var provider = undefined;                                  // 1753
            if (Router.current().route.getName() === "shared") {       // 1754
              provider = { rawParentToken: Router.current().params.token };
            } else {                                                   //
              provider = { identityId: senderGrain.identityId() };     // 1757
            }                                                          //
                                                                       //
            var owner = {                                              // 1760
              webkey: {                                                // 1761
                forSharing: forSharing,                                // 1762
                expiresIfUnusedDuration: selfDestructDuration          // 1763
              }                                                        //
            };                                                         //
                                                                       //
            var params = [provider, senderGrain.grainId(), petname, assignment, owner];
                                                                       //
            var memoizeKey = SHA256(JSON.stringify(params));           // 1769
            var memoizeResult = memoizedNewApiToken[memoizeKey];       // 1770
            if (memoizeResult && Date.now() - memoizeResult.timestamp > selfDestructDuration / 2) {
              // Memoized result is too old. Discard.                  //
              memoizeResult = undefined;                               // 1773
            }                                                          //
                                                                       //
            if (!memoizeResult) {                                      // 1776
              memoizedNewApiToken[memoizeKey] = memoizeResult = {      // 1777
                timestamp: Date.now(),                                 // 1778
                promise: new Promise(function (resolve, reject) {      // 1779
                  var callback = function (err, result) {              // 1780
                    if (err) {                                         // 1781
                      reject(err);                                     // 1782
                    } else {                                           //
                      resolve(result);                                 // 1784
                    }                                                  //
                  };                                                   //
                                                                       //
                  Meteor.call.apply(Meteor, ["newApiToken"].concat(params, callback));
                })                                                     //
              };                                                       //
            }                                                          //
                                                                       //
            memoizeResult.promise.then(function (result) {             // 1793
              var tokenId = result.token;                              // 1794
              // Generate random key id2.                              //
              var id2 = Random.secret();                               // 1796
              // Store apitoken id1 and template in session storage in the offer
              // template namespace under key id2.                     //
              var key = "offerTemplate" + id2;                         // 1799
              var host = globalDb.makeApiHost(tokenId);                // 1800
              var renderedTemplate = template.replace(/\$API_TOKEN/g, tokenId).replace(/\$API_HOST/g, host);
              sessionStorage.setItem(key, JSON.stringify({             // 1803
                token: tokenId,                                        // 1804
                renderedTemplate: renderedTemplate,                    // 1805
                expires: Date.now() + selfDestructDuration,            // 1806
                host: host                                             // 1807
              }));                                                     //
                                                                       //
              // Send message to event.source with URL containing id2  //
              templateLink = window.location.origin + "/offer-template.html#" + id2;
              event.source.postMessage({ rpcId: rpcId, uri: templateLink }, event.origin);
            }, function (error) {                                      //
              event.source.postMessage({ rpcId: rpcId, error: error.toString() }, event.origin);
            });                                                        //
          })();                                                        //
                                                                       //
          if (typeof _ret6 === "object") return _ret6.v;               //
        } else if (event.data.powerboxRequest) {                       //
          var powerboxRequest = event.data.powerboxRequest;            // 1818
          check(powerboxRequest, {                                     // 1819
            // TODO: be more strict, and check more fields, once the test apps are more conformant
            rpcId: Match.Any,                                          // 1821
            query: Match.Optional([Object]), // TODO: be more precise, and also make non-optional once tests are updated to provide a query
            saveLabel: Match.Optional(String)                          // 1823
          });                                                          //
                                                                       //
          var powerboxRequestInfo = {                                  // 1826
            source: event.source,                                      // 1827
            origin: event.origin,                                      // 1828
            // We'll need these to post a reply message                //
                                                                       //
            rpcId: powerboxRequest.rpcId,                              // 1831
            query: powerboxRequest.query,                              // 1832
            saveLabel: powerboxRequest.saveLabel,                      // 1833
            // These data come from the grain                          //
                                                                       //
            sessionId: senderGrain.sessionId(),                        // 1836
            grainId: senderGrain.grainId(),                            // 1837
            identityId: senderGrain.identityId(),                      // 1838
            // Attach grain context to the request.                    //
                                                                       //
            onCompleted: function () {                                 // 1841
              globalTopbar.closePopup();                               // 1841
            }                                                          //
          };                                                           //
          // Allow the grain to close the popup when we've completed the request.
          var requestContext = new SandstormPowerboxRequest(globalDb, powerboxRequestInfo);
          senderGrain.setPowerboxRequest(requestContext);              // 1845
          globalTopbar.openPopup("request");                           // 1846
        } else {                                                       //
          console.log("postMessage from app not understood: ", event.data);
          console.log(event);                                          // 1849
        }                                                              //
      };                                                               //
                                                                       //
      window.addEventListener("message", messageListener, false);      // 1853
    });                                                                //
  })();                                                                //
}                                                                      //
                                                                       //
if (Meteor.isClient) {                                                 // 1857
  function maybeScrollLog() {                                          // 1858
    var elem = document.getElementById("grainLog");                    // 1859
    if (elem) {                                                        // 1860
      // The log already exists. It's about to be updated. Check if it's scrolled to the bottom
      // before the update.                                            //
      if (elem.scrollHeight - elem.scrollTop === elem.clientHeight) {  // 1863
        // Indeed, so we want to scroll it back to the bottom after the update.
        Tracker.afterFlush(function () {                               // 1865
          scrollLogToBottom(elem);                                     // 1865
        });                                                            //
      }                                                                //
    } else {                                                           //
      // No element exists yet, but it's probably about to be created, in which case we definitely
      // want to scroll it.                                            //
      Tracker.afterFlush(function () {                                 // 1870
        var elem2 = document.getElementById("grainLog");               // 1871
        if (elem2) scrollLogToBottom(elem2);                           // 1872
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  function scrollLogToBottom(elem) {                                   // 1877
    elem.scrollTop = elem.scrollHeight;                                // 1878
  }                                                                    //
}                                                                      //
                                                                       //
function makeGrainIdActive(grainId) {                                  // 1882
  var grains = globalGrains.get();                                     // 1883
  for (var i = 0; i < grains.length; i++) {                            // 1884
    var grain = grains[i];                                             // 1885
    if (grain.grainId() === grainId) {                                 // 1886
      if (!grain.isActive()) {                                         // 1887
        grain.setActive(true);                                         // 1888
      }                                                                //
    } else {                                                           //
      if (grain.isActive()) {                                          // 1891
        grain.setActive(false);                                        // 1892
      }                                                                //
    }                                                                  //
  }                                                                    //
}                                                                      //
                                                                       //
function getActiveGrain(grains) {                                      // 1898
  var idx = activeGrainIndex(grains);                                  // 1899
  return idx == -1 ? undefined : grains[idx];                          // 1900
}                                                                      //
                                                                       //
function activeGrainIndex(grains) {                                    // 1903
  for (var i = 0; i < grains.length; i++) {                            // 1904
    if (grains[i].isActive()) {                                        // 1905
      return i;                                                        // 1906
    }                                                                  //
  }                                                                    //
                                                                       //
  return -1;                                                           // 1910
}                                                                      //
                                                                       //
function grainOriginToIndex(grains, origin) {                          // 1913
  for (var i = 0; i < grains.length; i++) {                            // 1914
    if (grains[i].origin() === origin) {                               // 1915
      return i;                                                        // 1916
    }                                                                  //
  }                                                                    //
                                                                       //
  return -1;                                                           // 1920
}                                                                      //
                                                                       //
function grainIdToIndex(grains, grainId) {                             // 1923
  for (var i = 0; i < grains.length; i++) {                            // 1924
    var grain = grains[i];                                             // 1925
    if (grains[i].grainId() === grainId) {                             // 1926
      return i;                                                        // 1927
    }                                                                  //
  }                                                                    //
                                                                       //
  return -1;                                                           // 1931
}                                                                      //
                                                                       //
function mapGrainStateToTemplateData(grainState) {                     // 1934
  var error = grainState.error();                                      // 1935
  var templateData = {                                                 // 1936
    grainId: grainState.grainId(),                                     // 1937
    active: grainState.isActive(),                                     // 1938
    title: grainState.title(),                                         // 1939
    error: error && error.message,                                     // 1940
    unauthorized: error && error.error == 403,                         // 1941
    appOrigin: grainState.origin(),                                    // 1942
    hasNotLoaded: !grainState.hasLoaded(),                             // 1943
    sessionId: grainState.sessionId(),                                 // 1944
    originalPath: grainState._originalPath,                            // 1945
    interstitial: grainState.shouldShowInterstitial(),                 // 1946
    token: grainState.token(),                                         // 1947
    viewInfo: grainState.viewInfo()                                    // 1948
  };                                                                   //
  return templateData;                                                 // 1950
}                                                                      //
                                                                       //
GrainLog = new Mongo.Collection("grainLog");                           // 1953
// Pseudo-collection created by subscribing to "grainLog", implemented in proxy.js.
                                                                       //
Router.map(function () {                                               // 1956
  this.route("apps", {                                                 // 1957
    path: "/apps",                                                     // 1958
    waitOn: function () {                                              // 1959
      return globalSubs;                                               // 1959
    },                                                                 //
                                                                       //
    data: function () {                                                // 1961
      if (!this.ready()) return;                                       // 1962
      if (!Meteor.userId() && !Meteor.loggingIn()) {                   // 1963
        Router.go("root", {}, { replaceState: true });                 // 1964
      }                                                                //
                                                                       //
      return new SandstormAppList(globalDb, globalQuotaEnforcer);      // 1967
    }                                                                  //
  });                                                                  //
  this.route("newGrainRedirect", {                                     // 1970
    // The path /grain/new used to be where you would go to create new grains.
    // Its functionality has been superceded by the apps route, so redirect in
    // case anyone has the old link saved somewhere.                   //
    path: "/grain/new",                                                // 1974
    onBeforeAction: function () {                                      // 1975
      Router.go("apps", {}, { replaceState: true });                   // 1976
    }                                                                  //
  });                                                                  //
  this.route("grains", {                                               // 1979
    path: "/grain",                                                    // 1980
    waitOn: function () {                                              // 1981
      return globalSubs;                                               // 1981
    },                                                                 //
                                                                       //
    data: function () {                                                // 1983
      if (!this.ready()) return;                                       // 1984
      if (!Meteor.userId() && !Meteor.loggingIn()) {                   // 1985
        Router.go("root", {}, { replaceState: true });                 // 1986
      }                                                                //
                                                                       //
      return new SandstormGrainListPage(globalDb, globalQuotaEnforcer);
    }                                                                  //
  });                                                                  //
  this.route("grain", {                                                // 1992
    path: "/grain/:grainId/:path(.*)?",                                // 1993
    loadingTemplate: "loadingNoMessage",                               // 1994
                                                                       //
    waitOn: function () {                                              // 1996
      return globalSubs;                                               // 1997
    },                                                                 //
                                                                       //
    onBeforeAction: function () {                                      // 2000
      // Run this hook only once.                                      //
      if (this.state.get("beforeActionHookRan")) {                     // 2002
        return this.next();                                            // 2002
      }                                                                //
                                                                       //
      this.state.set("beforeActionHookRan", true);                     // 2004
      var grainId = this.params.grainId;                               // 2005
      var initialPopup = null;                                         // 2006
      var shareGrain = Session.get("share-grain-" + grainId);          // 2007
      if (shareGrain) {                                                // 2008
        initialPopup = "share";                                        // 2009
      }                                                                //
                                                                       //
      var path = "/" + (this.params.path || "") + (this.originalUrl.match(/[#?].*$/) || "");
      var grains = globalGrains.get();                                 // 2013
      var grainIndex = grainIdToIndex(grains, grainId);                // 2014
      if (grainIndex == -1) {                                          // 2015
        // The element we need to attach our Blaze view to may not exist yet.
        // In that case, defer creating the GrainView until we're sure it's
        // had a chance to render.                                     //
        var openView = (function () {                                  // 2019
          function openView() {                                        // 2019
            var mainContentElement = document.querySelector("body>.main-content");
            if (mainContentElement) {                                  // 2021
              var _grains = globalGrains.get();                        // 2022
              var grainToOpen = new GrainView(grainId, path, undefined, mainContentElement, initialPopup);
              grainToOpen.openSession();                               // 2025
              grainIndex = _grains.push(grainToOpen) - 1;              // 2026
              globalGrains.set(_grains);                               // 2027
              makeGrainIdActive(grainId);                              // 2028
            } else {                                                   //
              Meteor.defer(openView);                                  // 2030
            }                                                          //
          }                                                            //
                                                                       //
          return openView;                                             //
        })();                                                          //
                                                                       //
        openView();                                                    // 2034
      } else {                                                         //
        makeGrainIdActive(grainId);                                    // 2036
      }                                                                //
                                                                       //
      this.next();                                                     // 2039
    },                                                                 //
                                                                       //
    onStop: function () {                                              // 2042
      this.state.set("beforeActionHookRan", undefined);                // 2043
      globalGrains.get().forEach(function (grain) {                    // 2044
        if (grain.isActive()) {                                        // 2045
          grain.setActive(false);                                      // 2046
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("/shared/:token/:path(.*)?", {                            // 2052
    name: "shared",                                                    // 2053
    template: "grain",                                                 // 2054
    loadingTemplate: "loadingNoMessage",                               // 2055
                                                                       //
    waitOn: function () {                                              // 2057
      return [Meteor.subscribe("devPackages"), Meteor.subscribe("tokenInfo", this.params.token), Meteor.subscribe("grainsMenu")];
    },                                                                 //
                                                                       //
    // This subscription gives us the data we need for deciding whether to automatically reveal
    // our identity.                                                   //
    // TODO(soon): Subscribe to contacts instead.                      //
    onBeforeAction: function () {                                      // 2069
      // Run this hook only once. We could accomplish the same thing by using the `onRun()` hook
      // and waiting for `this.ready()`, but for some reason that fails in the case when a user
      // logs in while visiting a /shared/ link.                       //
      if (this.state.get("beforeActionHookRan")) {                     // 2073
        return this.next();                                            // 2073
      }                                                                //
                                                                       //
      this.state.set("beforeActionHookRan", true);                     // 2075
                                                                       //
      var token = this.params.token;                                   // 2077
      var path = "/" + (this.params.path || "") + (this.originalUrl.match(/[#?].*$/) || "");
      var hash = this.params.hash;                                     // 2079
                                                                       //
      var tokenInfo = TokenInfo.findOne({ _id: token });               // 2081
      if (tokenInfo && tokenInfo.invalidToken) {                       // 2082
        this.state.set("invalidToken", true);                          // 2083
      } else {                                                         //
        (function () {                                                 //
          var grainId = tokenInfo.grainId;                             // 2085
          var grains = globalGrains.get();                             // 2086
          var grainIndex = grainIdToIndex(grains, grainId);            // 2087
          if (grainIndex == -1) {                                      // 2088
            var openView = (function () {                              // 2089
              function openView() {                                    // 2089
                var mainContentElement = document.querySelector("body>.main-content");
                if (mainContentElement) {                              // 2091
                  var _grains2 = globalGrains.get();                   // 2092
                  var grainToOpen = new GrainView(grainId, path, tokenInfo, mainContentElement);
                  grainToOpen.openSession();                           // 2094
                  grainIndex = _grains2.push(grainToOpen) - 1;         // 2095
                  globalGrains.set(_grains2);                          // 2096
                  makeGrainIdActive(grainId);                          // 2097
                } else {                                               //
                  Meteor.defer(openView);                              // 2099
                }                                                      //
              }                                                        //
                                                                       //
              return openView;                                         //
            })();                                                      //
                                                                       //
            openView();                                                // 2103
          } else {                                                     //
            makeGrainIdActive(grainId);                                // 2105
          }                                                            //
        })();                                                          //
      }                                                                //
                                                                       //
      this.next();                                                     // 2109
    },                                                                 //
                                                                       //
    action: function () {                                              // 2112
      if (this.state.get("invalidToken")) {                            // 2113
        this.render("invalidToken", { data: { token: this.params.token } });
      } else {                                                         //
        this.render();                                                 // 2116
      }                                                                //
    },                                                                 //
                                                                       //
    onStop: function () {                                              // 2120
      this.state.set("beforeActionHookRan", undefined);                // 2121
      this.state.set("invalidToken", undefined);                       // 2122
      globalGrains.get().forEach(function (grain) {                    // 2123
        if (grain.isActive()) {                                        // 2124
          grain.setActive(false);                                      // 2125
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("grainLog", {                                             // 2131
    path: "/grainlog/:grainId",                                        // 2132
    layoutTemplate: "noLayout",                                        // 2133
                                                                       //
    waitOn: function () {                                              // 2135
      return [Meteor.subscribe("grainTopBar", this.params.grainId), Meteor.subscribe("grainLog", this.params.grainId)];
    },                                                                 //
                                                                       //
    data: function () {                                                // 2142
      if (this.ready()) {                                              // 2143
        maybeScrollLog();                                              // 2144
        var grain = Grains.findOne(this.params.grainId);               // 2145
        return {                                                       // 2146
          title: grain ? grain.title : "(deleted grain)",              // 2147
          // jscs:disable requireCamelCaseOrUpperCaseIdentifiers       //
          html: AnsiUp.ansi_to_html(GrainLog.find({}, { $sort: { _id: 1 } }).map(function (entry) {
            return entry.text;                                         // 2150
          }).join(""), { use_classes: true })                          //
        };                                                             //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  // jscs:enable requireCamelCaseOrUpperCaseIdentifiers                //
  this.route("appDetails", {                                           // 2158
    path: "/apps/:appId",                                              // 2159
    template: "appDetails",                                            // 2160
    waitOn: function () {                                              // 2161
      return globalSubs.concat(Meteor.subscribe("appIndex", this.params.appId));
    },                                                                 //
                                                                       //
    data: function () {                                                // 2166
      return new SandstormAppDetails(globalDb, globalQuotaEnforcer, this.params.appId);
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("share", {                                                // 2171
    path: "/share/:grainId/:identityId",                               // 2172
    template: "share",                                                 // 2173
    waitOn: function () {                                              // 2174
      return Meteor.subscribe("grainTopBar", this.params.grainId);     // 2175
    },                                                                 //
                                                                       //
    data: function () {                                                // 2178
      var userId = Meteor.userId();                                    // 2179
      var grainId = this.params.grainId;                               // 2180
      var identityId = this.params.identityId;                         // 2181
      var grain = Grains.findOne({ _id: grainId });                    // 2182
      if (!grain) {                                                    // 2183
        return { grainNotFound: grainId };                             // 2184
      } else {                                                         //
        Session.set("share-grain-" + grainId, identityId);             // 2186
        Router.go("/grain/" + grainId);                                // 2187
      }                                                                //
    }                                                                  //
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=grain.js.map
