(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MeteorWrapperLdapjs = Package['typ:ldapjs'].MeteorWrapperLdapjs;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var ECMAScript = Package.ecmascript.ECMAScript;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var Future, LDAP_DEFAULTS, LDAP;

(function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                             //
// packages/accounts-ldap/ldap_server.js                                                                       //
//                                                                                                             //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                               //
Future = Npm.require("fibers/future");                                                                         // 1
                                                                                                               //
// At a minimum, set up LDAP_DEFAULTS.url and .dn according to                                                 //
// your needs. url should appear as 'ldap://your.url.here'                                                     //
// dn should appear in normal ldap format of comma separated attribute=value                                   //
// e.g. 'uid=someuser,cn=users,dc=somevalue'                                                                   //
LDAP_DEFAULTS = {                                                                                              // 7
  url: false,                                                                                                  // 8
  port: "389",                                                                                                 // 9
  dn: false,                                                                                                   // 10
  searchDN: false,                                                                                             // 11
  searchCredentials: false,                                                                                    // 12
  createNewUser: true,                                                                                         // 13
  defaultDomain: false,                                                                                        // 14
  searchResultsProfileMap: false,                                                                              // 15
  base: null,                                                                                                  // 16
  search: "(objectclass=*)",                                                                                   // 17
  ldapsCertificate: false                                                                                      // 18
};                                                                                                             //
                                                                                                               //
/**                                                                                                            //
 @class LDAP                                                                                                   //
 @constructor                                                                                                  //
 */                                                                                                            //
var LDAP = function (options) {                                                                                // 25
  // Set options                                                                                               //
  this.options = _.clone(LDAP_DEFAULTS);                                                                       // 27
                                                                                                               //
  // Make sure options have been set                                                                           //
  try {                                                                                                        // 30
    check(this.options.url, String);                                                                           // 31
    //check(this.options.dn, String);                                                                          //
  } catch (e) {                                                                                                //
    throw new Meteor.Error("Bad Defaults", "Options not set. Make sure to set LDAP_DEFAULTS.url and LDAP_DEFAULTS.dn!");
  }                                                                                                            //
                                                                                                               //
  // Because NPM ldapjs module has some binary builds,                                                         //
  // We had to create a wraper package for it and build for                                                    //
  // certain architectures. The package typ:ldap-js exports                                                    //
  // 'MeteorWrapperLdapjs' which is a wrapper for the npm module                                               //
  this.ldapjs = MeteorWrapperLdapjs;                                                                           // 41
};                                                                                                             //
                                                                                                               //
/**                                                                                                            //
 * Attempt to bind (authenticate) ldap                                                                         //
 * and perform a dn search if specified                                                                        //
 *                                                                                                             //
 * @method ldapCheck                                                                                           //
 *                                                                                                             //
 * @param {Object} options  Object with username, ldapPass and overrides for LDAP_DEFAULTS object.             //
 * Additionally the searchBeforeBind parameter can be specified, which is used to search for the DN            //
 * if not provided.                                                                                            //
 */                                                                                                            //
LDAP.prototype.ldapCheck = function (db, options) {                                                            // 54
                                                                                                               //
  var _this = this;                                                                                            // 56
                                                                                                               //
  options = options || {};                                                                                     // 58
                                                                                                               //
  if (options.hasOwnProperty("username") && options.hasOwnProperty("ldapPass")) {                              // 60
    var _ret = (function () {                                                                                  //
      _this.options.base = db.getLdapBase();                                                                   // 61
      _this.options.dn = db.getLdapDnPattern().replace("$USERNAME", options.username);                         // 62
      _this.options.searchBeforeBind = {                                                                       // 63
        uid: options.username                                                                                  // 64
      };                                                                                                       //
                                                                                                               //
      var ldapAsyncFut = new Future();                                                                         // 67
                                                                                                               //
      // Create ldap client                                                                                    //
      var fullUrl = _this.options.url + ":" + _this.options.port;                                              // 70
      var client = null;                                                                                       // 71
                                                                                                               //
      if (_this.options.url.indexOf("ldaps://") === 0) {                                                       // 73
        client = _this.ldapjs.createClient({                                                                   // 74
          url: fullUrl,                                                                                        // 75
          tlsOptions: {                                                                                        // 76
            ca: [_this.options.ldapsCertificate]                                                               // 77
          }                                                                                                    //
        });                                                                                                    //
      } else {                                                                                                 //
        client = _this.ldapjs.createClient({                                                                   // 81
          url: fullUrl                                                                                         // 82
        });                                                                                                    //
      }                                                                                                        //
                                                                                                               //
      client.on("error", function (err) {                                                                      // 86
        ldapAsyncFut["return"]({                                                                               // 87
          error: err                                                                                           // 88
        });                                                                                                    //
      });                                                                                                      //
                                                                                                               //
      // Slide @xyz.whatever from username if it was passed in                                                 //
      // and replace it with the domain specified in defaults                                                  //
      var emailSliceIndex = options.username.indexOf("@");                                                     // 94
      var username = undefined;                                                                                // 95
      var domain = _this.options.defaultDomain;                                                                // 96
                                                                                                               //
      // If user appended email domain, strip it out                                                           //
      // And use the defaults.defaultDomain if set                                                             //
      if (emailSliceIndex !== -1) {                                                                            // 100
        username = options.username.substring(0, emailSliceIndex);                                             // 101
        domain = domain || options.username.substring(emailSliceIndex + 1, options.username.length);           // 102
      } else {                                                                                                 //
        username = options.username;                                                                           // 104
      }                                                                                                        //
                                                                                                               //
      // If DN is provided, use it to bind                                                                     //
      if (!_this.options.base) {                                                                               // 108
        // Attempt to bind to ldap server with provided info                                                   //
        client.bind(_this.options.searchDN || _this.options.dn, _this.options.searchCredentials || options.ldapPass, function (err) {
          try {                                                                                                // 113
            if (err) {                                                                                         // 114
              // Bind failure, return error                                                                    //
              throw new Meteor.Error(err.code, err.message);                                                   // 116
            }                                                                                                  //
                                                                                                               //
            var handleSearchProfile = function (retObject, bindAfterSearch) {                                  // 119
              retObject.emptySearch = true;                                                                    // 120
                                                                                                               //
              // use dn if given, else use the base for the ldap search                                        //
              var searchBase = _this.options.dn || _this.options.base;                                         // 123
              var searchOptions = {                                                                            // 124
                scope: "sub",                                                                                  // 125
                sizeLimit: 1,                                                                                  // 126
                filter: _this.options.search                                                                   // 127
              };                                                                                               //
                                                                                                               //
              client.search(searchBase, searchOptions, function (err, res) {                                   // 130
                if (err) {                                                                                     // 131
                  ldapAsyncFut["return"]({                                                                     // 132
                    error: err                                                                                 // 133
                  });                                                                                          //
                  return;                                                                                      // 135
                }                                                                                              //
                                                                                                               //
                var found = false;                                                                             // 138
                res.on("searchEntry", function (entry) {                                                       // 139
                  found = true;                                                                                // 140
                  retObject.emptySearch = false;                                                               // 141
                  // Add entry results to return object                                                        //
                  retObject.searchResults = _.omit(entry.object, "userPassword");                              // 143
                                                                                                               //
                  if (bindAfterSearch) {                                                                       // 145
                    client.bind(retObject.searchResults.dn, options.ldapPass, function (err) {                 // 146
                      try {                                                                                    // 147
                        if (err) {                                                                             // 148
                          throw new Meteor.Error(err.code, err.message);                                       // 149
                        }                                                                                      //
                                                                                                               //
                        ldapAsyncFut["return"](retObject);                                                     // 152
                      } catch (e) {                                                                            //
                        ldapAsyncFut["return"]({                                                               // 154
                          error: e                                                                             // 155
                        });                                                                                    //
                      }                                                                                        //
                    });                                                                                        //
                  } else ldapAsyncFut["return"](retObject);                                                    //
                });                                                                                            //
                                                                                                               //
                res.on("end", function () {                                                                    // 162
                  if (!found) {                                                                                // 163
                    ldapAsyncFut["return"]({                                                                   // 164
                      error: new Meteor.Error(500, "No user found")                                            // 165
                    });                                                                                        //
                  }                                                                                            //
                });                                                                                            //
              });                                                                                              //
            };                                                                                                 //
                                                                                                               //
            var retObject = {                                                                                  // 173
              username: username,                                                                              // 174
              searchResults: null,                                                                             // 175
              email: domain ? username + "@" + domain : false,                                                 // 176
              dn: _this.options.dn                                                                             // 177
            };                                                                                                 //
                                                                                                               //
            if (_this.options.searchDN) {                                                                      // 180
              handleSearchProfile(retObject, true);                                                            // 181
            } else {                                                                                           //
              handleSearchProfile(retObject, false);                                                           // 183
            }                                                                                                  //
          } catch (e) {                                                                                        //
            ldapAsyncFut["return"]({                                                                           // 186
              error: e                                                                                         // 187
            });                                                                                                //
          }                                                                                                    //
        });                                                                                                    //
      }                                                                                                        //
      // DN not provided, search for DN and use result to bind                                                 //
      else {                                                                                                   //
          (function () {                                                                                       //
            // initialize result                                                                               //
            var retObject = {                                                                                  // 195
              username: username,                                                                              // 196
              email: domain ? username + "@" + domain : false,                                                 // 197
              emptySearch: true,                                                                               // 198
              searchResults: {}                                                                                // 199
            };                                                                                                 //
                                                                                                               //
            var filter = _this.options.search;                                                                 // 202
            Object.keys(_this.options.searchBeforeBind).forEach(function (searchKey) {                         // 203
              filter = "&" + filter + "(" + searchKey + "=" + _this.options.searchBeforeBind[searchKey] + ")";
            });                                                                                                //
                                                                                                               //
            var searchOptions = {                                                                              // 207
              scope: "sub",                                                                                    // 208
              sizeLimit: 1,                                                                                    // 209
              filter: filter                                                                                   // 210
            };                                                                                                 //
                                                                                                               //
            // perform LDAP search to determine DN                                                             //
            client.search(_this.options.base, searchOptions, function (err, res) {                             // 214
              if (err) {                                                                                       // 215
                ldapAsyncFut["return"]({                                                                       // 216
                  error: err                                                                                   // 217
                });                                                                                            //
                return;                                                                                        // 219
              }                                                                                                //
                                                                                                               //
              retObject.emptySearch = true;                                                                    // 222
              res.on("searchEntry", function (entry) {                                                         // 223
                retObject.dn = entry.objectName;                                                               // 224
                retObject.username = retObject.dn;                                                             // 225
                retObject.emptySearch = false;                                                                 // 226
                                                                                                               //
                retObject.searchResults = _.omit(entry.object, "userPassword");                                // 228
                                                                                                               //
                // use the determined DN to bind                                                               //
                client.bind(entry.objectName, options.ldapPass, function (err) {                               // 231
                  try {                                                                                        // 232
                    if (err) {                                                                                 // 233
                      throw new Meteor.Error(err.code, err.message);                                           // 234
                    } else {                                                                                   //
                      ldapAsyncFut["return"](retObject);                                                       // 236
                    }                                                                                          //
                  } catch (e) {                                                                                //
                    ldapAsyncFut["return"]({                                                                   // 240
                      error: e                                                                                 // 241
                    });                                                                                        //
                  }                                                                                            //
                });                                                                                            //
              });                                                                                              //
              // If no dn is found, return as is.                                                              //
              res.on("end", function (result) {                                                                // 247
                if (retObject.dn === undefined) {                                                              // 248
                  ldapAsyncFut["return"](retObject);                                                           // 249
                }                                                                                              //
              });                                                                                              //
            });                                                                                                //
          })();                                                                                                //
        }                                                                                                      //
                                                                                                               //
      return {                                                                                                 // 255
        v: ldapAsyncFut.wait()                                                                                 //
      };                                                                                                       //
    })();                                                                                                      //
                                                                                                               //
    if (typeof _ret === "object") return _ret.v;                                                               //
  } else {                                                                                                     //
    throw new Meteor.Error(403, "Missing LDAP Auth Parameter");                                                // 258
  }                                                                                                            //
};                                                                                                             //
                                                                                                               //
// Register login handler with Meteor                                                                          //
// Here we create a new LDAP instance with options passed from                                                 //
// Meteor.loginWithLDAP on client side                                                                         //
// @param {Object} loginRequest will consist of username, ldapPass, ldap, and ldapOptions                      //
Accounts.registerLoginHandler("ldap", function (loginRequest) {                                                // 267
  // If 'ldap' isn't set in loginRequest object,                                                               //
  // then this isn't the proper handler (return undefined)                                                     //
  if (!loginRequest.ldap) {                                                                                    // 270
    return undefined;                                                                                          // 271
  }                                                                                                            //
                                                                                                               //
  if (!Accounts.identityServices.ldap.isEnabled()) {                                                           // 274
    throw new Meteor.Error(403, "LDAP service is disabled.");                                                  // 275
  }                                                                                                            //
                                                                                                               //
  // Instantiate LDAP with options                                                                             //
  var userOptions = loginRequest.ldapOptions || {};                                                            // 279
  var ldapObj = new LDAP(userOptions);                                                                         // 280
                                                                                                               //
  // Call ldapCheck and get response                                                                           //
  var ldapResponse = ldapObj.ldapCheck(this.connection.sandstormDb, loginRequest);                             // 283
                                                                                                               //
  if (ldapResponse.error) {                                                                                    // 285
    return {                                                                                                   // 286
      userId: null,                                                                                            // 287
      error: ldapResponse.error                                                                                // 288
    };                                                                                                         //
  } else if (ldapResponse.emptySearch) {                                                                       //
    return {                                                                                                   // 291
      userId: null,                                                                                            // 292
      error: new Meteor.Error(403, "User not found in LDAP")                                                   // 293
    };                                                                                                         //
  } else {                                                                                                     //
    // Set initial userId and token vals                                                                       //
    return Accounts.updateOrCreateUserFromExternalService("ldap", { id: ldapResponse.dn, rawAttrs: ldapResponse.searchResults }, {});
  }                                                                                                            //
});                                                                                                            //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['accounts-ldap'] = {
  LDAP: LDAP,
  LDAP_DEFAULTS: LDAP_DEFAULTS
};

})();

//# sourceMappingURL=accounts-ldap.js.map
